import torch
from tqdm import tqdm
from torch import Tensor
from .eval_recog import valuate
from ..representation.eval_face import valuate as valuate_face
from ..representation.eval_cbir import valuate as valuate_cbir
from typing import Callable
import math
from ..optimizer import SAM
from torch.utils.tensorboard import SummaryWriter
import os
from copy import deepcopy

__all__ = ['Trainer']

def make_divisible(x: int, divisor = 32):
    # Returns nearest x divisible by divisor
    return math.ceil(x / divisor) * divisor

def print_imgsz(images: torch.Tensor):
    h, w = images.shape[-2:]
    return [h,w]

def mixup_data(x, y, device, lam):
    '''Returns mixed inputs, pairs of targets'''
    batch_size = x.size()[0]
    # to device
    index = torch.randperm(batch_size).to(device)

    mixed_x = lam * x + (1 - lam) * x[index, :]
    y_a, y_b = y, y[index]
    return mixed_x, y_a, y_b

def mixup_criterion(criterion, pred, y_a, y_b, lam):
    return lam * criterion(pred, y_a) + (1 - lam) * criterion(pred, y_b)

class Trainer:
    def __init__(self,
                 model,
                 train_dataloader,
                 val_dataloader,
                 optimizer,
                 scaler,
                 device: torch.device,
                 epochs: int,
                 logger,
                 rank: int,
                 scheduler,
                 ema, 
                 sampler = None,
                 thresh = 0,
                 teacher = None,
                 mixup_sampler = None,
                 # face
                 task: str = 'face',
                 print_freq = 50,
                 save_freq = 5,
                 cfgs: dict = None,
                 out_dir = None
                 ):

        self.model = model
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.optimizer = optimizer
        self.scaler = scaler
        self.device = device
        self.epochs = epochs
        self.logger = logger
        self.rank = rank
        self.scheduler = scheduler
        self.ema = ema
        self.sampler = sampler
        self.thresh = thresh
        self.teacher = teacher
        self.sam: bool = type(self.optimizer) is SAM
        self.distill: bool = teacher is not None
        self.mixup_sampler = mixup_sampler

        # face
        self.task = task
        self.print_freq = print_freq
        self.save_freq = save_freq
        self.data_cfg = cfgs['data']
        self.model_cfg = cfgs['model']
        self.hyp_cfg = cfgs['hyp']
        if rank in (-1, 0) and out_dir is not None:
            self.writer = SummaryWriter(log_dir=out_dir)

    def train_one_epoch(self, epoch: int, criterion: Callable):
        # train mode
        self.model.train()

        cuda: bool = self.device != torch.device('cpu')

        if self.rank != -1:
            self.train_dataloader.sampler.set_epoch(epoch)
        pbar = enumerate(self.train_dataloader)
        if self.rank in {-1, 0}:
            pbar = tqdm(enumerate(self.train_dataloader),
                        total=len(self.train_dataloader),
                        bar_format='{l_bar}{bar:10}{r_bar}')

        tloss, fitness = 0., 0.

        for i, (images, labels) in pbar:  # progress bar

            if self.mixup_sampler is not None:
                lam = self.mixup_sampler.sample()
            else: 
                lam = 0

            images, labels = images.to(self.device, non_blocking=True), labels.to(self.device)
            if self.sampler is not None:  # OHEM-Softmax
                with torch.no_grad():
                    valid = self.sampler.sample(self.model(images), labels)
                    images, labels = images[valid], labels[valid]
            with torch.autocast(device_type=self.device.type, enabled=(self.device != torch.device('cpu'))):
                loss = self.compute_loss(images, labels, lam, criterion)

            if self.rank in {-1, 0}:
                tloss = (tloss * i + loss.item()) / (i + 1)  # update mean losses
                mem = '%.3gG' % (torch.cuda.memory_reserved() / 1E9 if cuda else 0)  # (GB)
                pbar.desc = f"{f'{epoch + 1}/{self.epochs}':>10}{mem:>10}{tloss:>12.3g}" + ' ' * 36
                pbar.postfix = f'lr:{self.optimizer.param_groups[0]["lr"]:.5f}, imgsz:{print_imgsz(images)}'

                if i == len(pbar) - 1:  # last batch
                    self.logger.log(f'EPOCH:{epoch + 1:d}  Train-Loss:{tloss:4f}  LR:{self.optimizer.param_groups[0]["lr"]:.5f}')
                    if self.thresh == 0:
                        # val
                        top1, top5, v_loss = valuate(self.ema.ema, self.val_dataloader, self.device, pbar, True, criterion, self.logger,
                                                     self.thresh)
                        self.logger.log(f'VAL-LOSS:{v_loss:4f}\n')
                    else:
                        self.logger.log(f'{"name":<8}{"nums":>8}{"precision":>15}{"recall":>10}{"f1score":>10}')
                        # val
                        precision, recall, f1score, v_loss = valuate(self.ema.ema, self.val_dataloader, self.device, pbar, True,
                                                                     criterion, self.logger, self.thresh)
                        self.logger.log(f'VAL-Loss:{v_loss:4f}\n')

                    fitness = top1 if self.thresh == 0 else f1score  # define fitness as top1 accuracy

        self.scheduler.step()  # step epoch-wise

        return fitness

    @staticmethod
    def update_sam(model: torch.nn.Module, inputs, targets, optimizer, lossfn, rank, ema=None, mixup=False, **kwargs):
        # first forward-backward step
        optimizer.enable_running_stats(model)
        if not mixup:
            loss = lossfn(model(inputs), targets)
        else:
            loss = mixup_criterion(lossfn, model(inputs), **kwargs)
        if rank >= 0:  # multi-gpu
            with model.no_sync():
                loss.mean().backward()
        else:
            loss.mean().backward()
        optimizer.first_step(zero_grad=True)

        # second forward-backward step
        optimizer.disable_running_stats(model)
        if not mixup:
            lossfn(model(inputs), targets).mean().backward()
        else:
            mixup_criterion(lossfn, model(inputs), **kwargs).mean().backward()
        optimizer.second_step(zero_grad=True)

        if ema:
            ema.update(model)

        return loss

    def compute_loss(self, images: Tensor, labels: Tensor, lam: float, criterion: Callable, face: bool = False):
        mixup: bool = lam > 0

        assert not (mixup and self.distill), 'distill not be True when mixup is True'
        if mixup and self.sam: # close
            images, targets_a, targets_b = mixup_data(images, labels, self.device, lam)
            kwargs = dict(y_a=targets_a, y_b=targets_b, lam=lam)
            loss = Trainer.update_sam(self.model, images, labels, self.optimizer, criterion, self.rank, self.ema, mixup=True, **kwargs)
        elif mixup: # close
            images, targets_a, targets_b = mixup_data(images, labels, self.device, lam)
            loss = mixup_criterion(criterion, self.model(images), targets_a, targets_b, lam)
            Trainer.update(self.model, loss, self.scaler, self.optimizer, self.ema)
        elif self.sam and self.distill:
            raise ValueError('SAM optimizer and Knowledge distilling have not been implemented yet.')
        elif self.sam: # close
            loss = Trainer.update_sam(self.model, images, labels, self.optimizer, criterion, self.rank, self.ema, mixup=False)
        elif self.distill:
            raise ValueError('Knowledge distilling have not been implemented yet.')
        else: # close
            loss = criterion(self.model(images), labels) if not face else criterion(self.model(images, labels), labels)
            Trainer.update(self.model, loss, self.scaler, self.optimizer, self.ema)

        return loss

    # scale + backward + grad_clip + step + zero_grad
    @staticmethod
    def update(model, loss, scaler, optimizer, ema=None):
        # backward
        scaler.scale(loss).backward()

        # optimize
        scaler.unscale_(optimizer)  # unscale gradients
        torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=10.0)  # clip gradients
        scaler.step(optimizer)
        scaler.update()

        optimizer.zero_grad()
        if ema:
            ema.update(model)

    def train_one_epoch_face(self, criterion, cur_epoch, loss_meter):
        """Tain one epoch by traditional training.
        """

        # config
        config = {}
        config["model"] = self.model_cfg
        config["data"] = self.data_cfg
        config["hyp"] = self.hyp_cfg

        self.model.train()
        iters_per_epoch = len(self.train_dataloader)

        for batch_idx, (images, labels) in enumerate(self.train_dataloader):
            images, labels = images.to(self.device, non_blocking=True), labels.to(self.device)

            loss = self.compute_loss(images, labels, lam=0, criterion = criterion, face = True)

            global_batch_idx = cur_epoch * iters_per_epoch + batch_idx
            self.scheduler.step() # step batch-wise
            
            if self.rank in (-1, 0):
                loss_meter.update(loss.item(), images.shape[0])

            if self.rank in (-1, 0) and batch_idx % self.print_freq == 0:
                loss_avg = loss_meter.avg
                lr = self.optimizer.param_groups[0]["lr"]
                self.logger.both('Epoch %d, iter %d/%d, lr %f, loss %f' %
                            (cur_epoch+1, batch_idx+1, iters_per_epoch, lr, loss_avg))
                self.writer.add_scalar('Train_loss', loss_avg, global_batch_idx)
                self.writer.add_scalar('Train_lr', lr, global_batch_idx)
                loss_meter.reset()

            if self.rank in (-1, 0) and (cur_epoch * iters_per_epoch + batch_idx + 1) % (self.save_freq * iters_per_epoch)== 0:
                saved_name = 'Epoch_%d.pt' % (cur_epoch+1)
                if self.task == 'face':

                    mean, std = valuate_face(self.ema.ema.trainingwrapper['backbone'],
                                            self.data_cfg,
                                            self.device)
                    self.writer.add_scalar('Val_mean', mean, global_batch_idx)
                    self.writer.add_scalar('Val_std', std, global_batch_idx)

                    fitness = {'fitness': {'Val_mean': float(mean), 'Val_std': float(std)}}
                elif self.task == 'cbir':
                    metrics = valuate_cbir(self.ema.ema.trainingwrapper['backbone'],
                                           self.data_cfg,
                                           self.device,
                                           self.logger)
                    for k, v in metrics.items():
                        self.writer.add_scalar(f'Val_{k}', v, global_batch_idx)
                    fitness = {'fitness': metrics}
                
                fitness['checkpoint'] = saved_name
                    
                ckpt = {
                    'epoch': cur_epoch,
                    'batch_id': batch_idx,
                    'fitness': fitness,
                    'state_dict': self.model.trainingwrapper['backbone'].state_dict() if self.rank == -1 else self.model.module.trainingwrapper['backbone'].state_dict(),
                    'ema': deepcopy(self.ema.ema.trainingwrapper['backbone'].state_dict()),
                    'updates': self.ema.updates,
                    'optimizer': self.optimizer.state_dict(),  # optimizer.state_dict(),
                    'scheduler': self.scheduler.state_dict(),
                    'config': config
                }
                if self.device != torch.device('cpu'): ckpt['scaler'] = self.scaler.state_dict()

                torch.save(ckpt, os.path.join(self.writer.log_dir, saved_name))
                self.logger.both(fitness)
            torch.cuda.empty_cache()


from .basedataset import ImageDatasets
from ..built.class_augmenter import ClassWiseAugmenter
import torch
import os
from torch.utils.data import DataLoader
from typing import Optional

class SmartDataProcessor:
    def __init__(self, data_cfgs: dict, rank, project, training: bool = True):
        self.data_cfgs = data_cfgs # root, nw, imgsz, train, val
        self.rank = rank
        self.project = project
        self.label_transforms = None # used in CenterProcessor.__init__

        if training:
            self.train_dataset = self.create_dataset('train')

    def create_dataset(self, mode: str, training: bool = True, id2label: Optional[dict] = None):
        assert mode in {'train', 'val'}

        cfg = self.data_cfgs.get(mode, -1)
        if isinstance(cfg, dict):
            dataset = ImageDatasets(root_or_dataset=self.data_cfgs['root'], mode=mode,
                                    transforms=ClassWiseAugmenter(cfg['augment'], None, None) if mode == 'val' else \
                               ClassWiseAugmenter(cfg['augment'], cfg['class_aug'], cfg['base_aug']),
                                    project=self.project, rank=self.rank, training = training, id2label=id2label)
        else: 
            dataset = None
        return dataset

    def set_augment(self, mode: str, transforms = None): # sequence -> T.Compose([...])
        if transforms is None:
            transforms = self.val_dataset.transforms.base_transforms
        dataset = getattr(self, f'{mode}_dataset')
        dataset.transforms.base_transforms = transforms

    def auto_aug_weaken(self, epoch: int, milestone: int, sequence: Optional[torch.nn.Module] = None):
        if epoch == milestone:
            # sequence = create_AugTransforms('random_horizonflip to_tensor normalize')
            self.set_augment('train', transforms = sequence)

    @staticmethod
    def set_dataloader(dataset, bs: int = 256, nw: int = 0, pin_memory: bool = True, shuffle: bool = True, sampler = None, collate_fn= None, *args, **kwargs):
        assert not (shuffle and sampler is not None)
        nd = torch.cuda.device_count()
        nw = min([os.cpu_count() // max(nd, 1), nw])
        return DataLoader(dataset=dataset, batch_size=bs, num_workers=nw, pin_memory=pin_memory, sampler=sampler, shuffle=shuffle, collate_fn=collate_fn, *args, **kwargs)


import math
import torch
import torch.nn.functional as F
from torch.nn import Module, Parameter

class MagFace(Module):
    """Implementation for "ArcFace: Additive Angular Margin Loss for Deep Face Recognition"
    """
    def __init__(self, feat_dim, num_class, margin_am=0.0, scale=32, l_a=10, u_a=110, l_margin=0.45, u_margin=0.8, lamda=20):
        super(MagFace, self).__init__()
        self.weight = Parameter(torch.Tensor(feat_dim, num_class))
        self.weight.data.uniform_(-1, 1).renorm_(2, 1, 1e-5).mul_(1e5)
        self.margin_am = margin_am
        self.scale = scale        
        self.l_a = l_a
        self.u_a = u_a
        self.l_margin = l_margin
        self.u_margin = u_margin
        self.lamda = lamda

    def calc_margin(self, x):
        margin = (self.u_margin-self.l_margin) / \
                 (self.u_a-self.l_a)*(x-self.l_a) + self.l_margin
        return margin
    
    def forward(self, feats, labels):
        x_norm = torch.norm(feats, dim=1, keepdim=True).clamp(self.l_a, self.u_a)# l2 norm
        ada_margin = self.calc_margin(x_norm)
        cos_m, sin_m = torch.cos(ada_margin), torch.sin(ada_margin)
        loss_g = 1/(self.u_a**2) * x_norm + 1/(x_norm)
        kernel_norm = F.normalize(self.weight, dim=0)
        feats = F.normalize(feats)
        cos_theta = torch.mm(feats, kernel_norm) 
        cos_theta = cos_theta.clamp(-1, 1)
        sin_theta = torch.sqrt(1.0 - torch.pow(cos_theta, 2))
        cos_theta_m = cos_theta * cos_m - sin_theta * sin_m
        # 0 <= theta + m <= pi, ==> -m <= theta <= pi-m
        # because 0<=theta<=pi, so, we just have to keep theta <= pi-m, that is cos_theta >= cos(pi-m)
        min_cos_theta = torch.cos(math.pi - ada_margin)        
        cos_theta_m = torch.where(cos_theta > min_cos_theta, cos_theta_m, cos_theta-self.margin_am)
        index = torch.zeros_like(cos_theta)
        index.scatter_(1, labels.data.view(-1, 1), 1)
        index = index.byte().bool()
        output = cos_theta * 1.0
        output[index] = cos_theta_m[index]
        output *= self.scale
        return output, self.lamda*loss_g



import timm
import torch
import torch.nn as nn

class TimmWrapper(nn.Module):
    """A wrapper for timm models that handles different model architectures uniformly."""
    
    def __init__(self, 
                 model_name: str,
                 feat_dim: int,
                 image_size: int,
                 pretrained: bool = True,
                 **kwargs):
        super().__init__()
        
        self.model = timm.create_model(
            model_name,
            pretrained=pretrained,
            num_classes=0,  # classification head -> nn.Identity()
            global_pool='',  # global pooling -> nn.Identity()
        )
        
        with torch.no_grad():
            dummy_input = torch.zeros(1, 3, image_size, image_size)
            output = self.model(dummy_input)
            
            if isinstance(output, tuple):
                output = output[0]
            
            if len(output.shape) == 4:  # CNN output: [B, C, H, W]
                _, channels, h, w = output.shape
                flatten_dim = channels * h * w
                self.output_layer = nn.Sequential(
                    nn.BatchNorm2d(channels),
                    nn.Flatten(1),
                    nn.Linear(flatten_dim, feat_dim),
                    nn.BatchNorm1d(feat_dim)
                )
            elif len(output.shape) == 3:  # Transformer output: [B, N, C]
                _, tokens, channels = output.shape
                flatten_dim = tokens * channels
                self.output_layer = nn.Sequential(
                    nn.LayerNorm(channels),
                    nn.Flatten(1),
                    nn.Linear(flatten_dim, feat_dim),
                    nn.BatchNorm1d(feat_dim)
                )
            else:
                raise ValueError(f"Unexpected output shape: {output.shape}")

    def forward(self, x):
        x = self.model(x)
        x = self.output_layer(x)
        return x


from torch import Tensor
import torch
import torch.nn.functional as F

class OHEMImageSampler:
    def __init__(self, min_kept: int, thresh: float, ignore_index: int = 255):
        self.min_kept = min_kept
        self.thresh = thresh
        self.ignore_index = ignore_index

    def sample(self, logits: Tensor, labels: Tensor):
        with torch.no_grad():
            prob = F.softmax(logits, dim=1)
            # ignore ignore_index here
            valid1 = labels != self.ignore_index
            prob = prob[valid1]

            # extract the score for correct predictions
            tmp_prob = prob.gather(1, labels[valid1].unsqueeze(1)).squeeze(1)
            sort_prob, sort_indices = tmp_prob.sort()

            min_thresh = sort_prob[min(self.min_kept, sort_prob.numel()-1)]
            threshold = max(min_thresh, self.thresh)

            temp_valid = sort_prob < threshold
            valid_indices = sort_indices[temp_valid]

            valid2 = torch.zeros_like(labels, dtype=torch.bool)
            valid2[valid_indices] = True

            return valid1 & valid2



import torch
import torch.nn.functional as F
from torch.nn import Module, Parameter

class CircleLoss(Module):
    """Implementation for "Circle Loss: A Unified Perspective of Pair Similarity Optimization"
    Note: this is the classification based implementation of circle loss.
    """
    def __init__(self, feat_dim, num_class, margin=0.25, gamma=256):
        super(CircleLoss, self).__init__()
        self.weight = Parameter(torch.Tensor(feat_dim, num_class))
        self.weight.data.uniform_(-1, 1).renorm_(2, 1, 1e-5).mul_(1e5)
        self.margin = margin
        self.gamma = gamma

        self.O_p = 1 + margin
        self.O_n = -margin
        self.delta_p = 1-margin
        self.delta_n = margin

    def forward(self, feats, labels):
        kernel_norm = F.normalize(self.weight, dim=0)
        feats = F.normalize(feats)
        cos_theta = torch.mm(feats, kernel_norm) 
        cos_theta = cos_theta.clamp(-1, 1)
        index_pos = torch.zeros_like(cos_theta)        
        index_pos.scatter_(1, labels.data.view(-1, 1), 1)
        index_pos = index_pos.byte().bool()
        index_neg = torch.ones_like(cos_theta)        
        index_neg.scatter_(1, labels.data.view(-1, 1), 0)
        index_neg = index_neg.byte().bool()

        alpha_p = torch.clamp_min(self.O_p - cos_theta.detach(), min=0.)
        alpha_n = torch.clamp_min(cos_theta.detach() - self.O_n, min=0.)

        logit_p = alpha_p * (cos_theta - self.delta_p)
        logit_n = alpha_n * (cos_theta - self.delta_n)

        output = cos_theta * 1.0
        output[index_pos] = logit_p[index_pos]
        output[index_neg] = logit_n[index_neg]
        output *= self.gamma
        return output



import glob
from ...utils.plots import Annotator
import platform
import shutil
import os
import torch.nn.functional as F
import cv2
from typing import Union
import matplotlib.pyplot as plt
import numpy as np
from functools import partial

class Visualizer:
    
    @staticmethod
    def predict_images(model, 
                       dataloader, 
                       device, 
                       visual_path, 
                       class_indices: dict, 
                       logger, 
                       thresh: Union[float, list[float]], 
                       infer_option: str = 'default',
                       ):
        """
        Args:
            infer_option: 
                - 'default': Infer + Visualize + GradCAM + Badcase
                - 'autolabel': Infer + Label
        """
        os.makedirs(visual_path, exist_ok=True)
        is_single_label = isinstance(thresh, (int, float)) and thresh == 0
        
        # Determine classification head type and activation function once
        class_head = 'ce' if is_single_label else 'bce'
        activation_fn = partial(F.softmax, dim=0) if class_head == 'ce' else partial(F.sigmoid)

        target_classes = dataloader.dataset.classes if dataloader.dataset.classes is not None else None
        if target_classes and not isinstance(target_classes, list):
            target_classes = [target_classes]
            
        # 获取每个目标类别的索引和阈值
        target_indices = []
        target_thresholds = []
        if not is_single_label and isinstance(thresh, list):
            for target_class in target_classes:
                target_idx = None
                for idx, class_name in class_indices.items():
                    if class_name == target_class:
                        target_idx = idx
                        target_indices.append(idx)
                        break
                if target_idx is None:
                    raise ValueError(f"Target class {target_class} not found in class indices")
                
                # 获取并验证目标类别的阈值
                target_thresh = thresh[target_idx]
                if not isinstance(target_thresh, float):
                    raise ValueError(f"Invalid threshold type for target class: {type(target_thresh)}. Must be float")
                target_thresholds.append(target_thresh)
                
        # Initialize CAM if in default mode
        if infer_option == 'default':
            from ...utils.cam import ClassActivationMaper
            cam = ClassActivationMaper(model, method='gradcam', device=device, transforms=dataloader.dataset.transforms)

        # eval mode
        model.eval()
        n = len(dataloader)

        fixed_class_length = 15
        progress_width = len(str(n))

        image_postfix_table = dict()
        for i, (img, inputs, img_path, gt_labels) in enumerate(dataloader):
            img = img[0]
            img_path = img_path[0]
            gt_label = gt_labels[0] if gt_labels is not None else None

            if infer_option == 'default':
                cam_image = cam(image=img, input_tensor=inputs, dsize=img.size)
                cam_image = cv2.resize(cam_image, img.size, interpolation=cv2.INTER_LINEAR)

            # system
            if platform.system().lower() == 'windows':
                annotator = Annotator(img, font=r'C:/WINDOWS/FONTS/SIMSUN.TTC') # windows
            else:
                annotator = Annotator(img) # linux

            # transforms
            inputs = inputs.to(device)
            # forward
            logits = model(inputs).squeeze()

            # post process using pre-determined activation function
            probs = activation_fn(logits)
            top5i = probs.argsort(0, descending=True)[:5].tolist()
            
            text = '\n'.join(f'{class_indices[j]:<{fixed_class_length}} {probs[j].item():.2f}' for j in top5i)
            
            formatted_predictions = '      '.join(f'{class_indices[j]:<{fixed_class_length}}{probs[j].item():.2f}' for j in top5i)
            logger.console(f"[{i+1:>{progress_width}}|{n:<{progress_width}}] {os.path.basename(img_path):<20} {formatted_predictions}")

            annotator.text((32, 32), text, txt_color=(0, 0, 0))

            # Save predictions and ground truth
            save_dir = os.path.join(visual_path, 'labels')
            os.makedirs(save_dir, exist_ok=True)
            image_postfix_table[os.path.basename(os.path.splitext(img_path)[0] + '.txt')] = {
                'ext': os.path.splitext(img_path)[1],
                'gt': gt_label
            }
            with open(os.path.join(save_dir, os.path.basename(os.path.splitext(img_path)[0] + '.txt')), 'a') as f:
                f.write(text + '\n')

            if infer_option == 'default':
                img = np.hstack([cv2.cvtColor(np.array(img), cv2.COLOR_RGB2BGR), cam_image])
                cv2.imwrite(os.path.join(visual_path, os.path.basename(img_path)), img)

        # Process badcases only in default mode
        if infer_option == 'default':
            badcase_root = os.path.join(visual_path, 'badcase')
            if target_classes:
                for target_class in target_classes:
                    os.makedirs(os.path.join(badcase_root, target_class), exist_ok=True)
            else:
                os.makedirs(badcase_root, exist_ok=True)

            for txt in glob.glob(os.path.join(visual_path, 'labels', '*.txt')):
                with open(txt, 'r') as f:
                    lines = f.readlines()
                    gt = image_postfix_table[os.path.basename(txt)]['gt']
                    
                    if gt is None:
                        continue
                    
                    if is_single_label:
                        # 单标签情况保持不变
                        pred_class = lines[0].split()[0]
                        is_badcase = pred_class != gt
                        target_class = gt  # 用于确定保存路径
                    else:
                        # 多标签情况，检查每个目标类别
                        is_badcase = False
                        target_class = None  # 用于确定保存路径
                        for target, thresh in zip(target_classes, target_thresholds):
                            found_correct_pred = False
                            for line in lines:
                                class_name, prob = line.split()[0], float(line.split()[1])
                                if class_name == target:
                                    if prob < thresh:
                                        is_badcase = True
                                        target_class = target  # 记录导致badcase的类别
                                    found_correct_pred = True
                                    break
                            if not found_correct_pred:
                                is_badcase = True
                                target_class = target
                
                if is_badcase:
                    try:
                        source_path = os.path.join(visual_path, 
                                                 os.path.basename(txt).replace('.txt', 
                                                 image_postfix_table[os.path.basename(txt)]['ext']))
                        if target_class and target_classes:
                            dest_path = os.path.join(badcase_root, target_class)
                        else:
                            dest_path = badcase_root
                        shutil.move(source_path, dest_path)
                    except FileNotFoundError:
                        print(f'FileNotFoundError->{txt}')

    @staticmethod
    def visualize_results(query, 
                          retrieval_results, 
                          scores, 
                          ground_truths, 
                          savedir,
                          max_rank=5,
                          query_dataset=None,
                          gallery_dataset=None
                          ):

        os.makedirs(savedir, exist_ok=True)

        fig, axes = plt.subplots(2, max_rank + 1, figsize=(3 * (max_rank + 1), 12))

        for ax in axes.ravel():
            ax.set_axis_off()
        # Display the query image in the first position of the second row
        query_img = query_dataset.get_image(query)
        ax = fig.add_subplot(2, max_rank + 1, max_rank + 2)
        ax.imshow(query_img)
        ax.set_title('Query')
        ax.axis("off")

        # Display the ground truth images
        for i in range(min(5, len(ground_truths))):
            gt_img = gallery_dataset.get_image(ground_truths[i])
            ax = fig.add_subplot(2, max_rank + 1, i + 1)
            ax.imshow(gt_img)
            ax.set_title('Ground Truth')
            ax.axis("off")

        # Display the retrieval images
        for i in range(max_rank):
            retrieval_img = gallery_dataset.get_image(retrieval_results[i])

            score = scores[i]
            is_tp = retrieval_results[i] in ground_truths
            label = 'true' if is_tp else 'false'
            color = (1, 0, 0)

            ax = fig.add_subplot(2, max_rank + 1, (max_rank + 1) + i + 2)
            if is_tp:
                ax.add_patch(plt.Rectangle(xy=(0, 0), width=retrieval_img.width - 1,
                                           height=retrieval_img.height - 1, edgecolor=color,
                                           fill=False, linewidth=8))
            ax.imshow(retrieval_img)
            ax.set_title('{:.4f}/{}'.format(score, label))
            ax.axis("off")

        #plt.tight_layout()
        image_id = os.path.basename(os.path.dirname(query))
        image_name = os.path.basename(query)
        image_unique = image_id + '_' + image_name
        fig.savefig(os.path.join(savedir, image_unique))
        plt.close(fig)


import logging
from typing import Union
import yaml

class SmartLogger:

    _Instance = None
    _Flag = False

    def __new__(cls, *args, **kwargs):
        if cls._Instance is None:
            cls._Instance = super().__new__(cls)
        return cls._Instance

    def __init__(self, filename = None, level: int = 1):
        if not self.__class__._Flag:
            self.__class__._Flag = True
            # logger -> Singleton
            self.file_logger = self.create_logger('file', 'file', filename=filename) if filename is not None else None
            self.console_logger = self.create_logger('console', 'console')
            self.level = level

    def log(self, msg: Union[str, dict]):
        if isinstance(msg, dict):
            self.file_logger.info(yaml.dump(msg, sort_keys=False, default_flow_style=False))

        else: self.file_logger.info(msg)

    def console(self, msg: Union[str, dict]):
        if isinstance(msg, dict):
            self.console_logger.info('\n'+str(yaml.dump(msg, sort_keys=False, default_flow_style=False)))

        else:
            self.console_logger.info(msg)

    def both(self, msg: Union[str, dict]):
        self.log(msg)
        self.console(msg)

    def create_logger(self, name: str, kind: str, filename = None):
        assert kind in {'file', 'console'}
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        # format
        file_format = logging.Formatter(fmt='%(asctime)-20s%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        if kind == 'file':
            handler = logging.FileHandler(filename=filename)
            handler.setFormatter(file_format)
            logger.addHandler(handler)
        elif kind == 'console':
            handler = logging.StreamHandler()
            handler.setFormatter(file_format)
            logger.addHandler(handler)
        return logger












from typing import Callable
from functools import wraps
from torch.optim.lr_scheduler import LinearLR, CosineAnnealingLR, SequentialLR

__all__ = ['linear',  # 线性衰减
           'cosine',  # 余弦衰减
           'linear_with_warm', # 线性衰减带热身
           'cosine_with_warm', # 余弦衰减带热身
           'create_Scheduler',
           'list_schedulers', ]

SCHEDULER = {}
def register_scheduler(fn: Callable):
    key = fn.__name__
    if key in SCHEDULER:
        raise ValueError(f"An entry is already registered under the name '{key}'.")
    SCHEDULER[key] = fn
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    return wrapper

def de_lrf_ratio(lrf_ratio):
    return 0.1 if lrf_ratio is None else lrf_ratio

@register_scheduler
def linear(optimizer, warm_ep, epochs, lr0, lrf_ratio):
    return LinearLR(optimizer, start_factor=1, end_factor=de_lrf_ratio(lrf_ratio), total_iters=epochs)

@register_scheduler
def cosine(optimizer, warm_ep, epochs, lr0, lrf_ratio):
    return CosineAnnealingLR(optimizer, T_max=epochs, eta_min=de_lrf_ratio(lrf_ratio) * lr0, )

@register_scheduler
def linear_with_warm(optimizer, warm_ep, epochs, lr0, lrf_ratio):
    scheduler = SequentialLR(
        optimizer = optimizer,
        schedulers=[
            LinearLR(optimizer, start_factor=0.1, end_factor=1, total_iters=warm_ep),
            LinearLR(optimizer, start_factor=1, end_factor=de_lrf_ratio(lrf_ratio), total_iters=epochs-warm_ep),
        ],
        milestones=[warm_ep,]
    )
    return scheduler

@register_scheduler
def cosine_with_warm(optimizer, warm_ep, epochs, lr0, lrf_ratio):
    scheduler = SequentialLR(
        optimizer=optimizer,
        schedulers=[
            LinearLR(optimizer, start_factor=0.1, end_factor=1, total_iters=warm_ep),
            CosineAnnealingLR(optimizer, T_max=epochs-warm_ep, eta_min=de_lrf_ratio(lrf_ratio) * lr0, )
        ],
        milestones=[warm_ep, ]
    )
    return scheduler

def create_Scheduler(scheduler, optimizer, warm_ep, epochs, lr0, lrf_ratio):
    return SCHEDULER[scheduler](optimizer, warm_ep, epochs, lr0, lrf_ratio)

def list_schedulers():
    lossfns = [k for k, v in SCHEDULER.items()]
    return sorted(lossfns)


from ..engine.optimizer import BaseSeperateLayer
from typing import Iterator, List, Dict, Union
from torch.nn import Module

class SeperateLayerParams(BaseSeperateLayer):
    def __init__(self, model: Module):
        super().__init__(model)

    def create_ParamSequence(self, layer_wise: bool, lr: float) -> Union[Iterator, List[Dict]]:
        """
        Args:
            layer_wise: lr衰减系数
            lr: 基准学习率

        Returns:
            params: torch.optim.Optimizer中的params
        """
        if not layer_wise: return self.model.parameters()

        # params = [
        #     {'params': self.model.features.parameters()},
        #     {'params': self.model.norm.parameters()},
        #     {'params': self.model.head.parameters(), 'lr': lr * 10}
        # ]

        params = [
            {'params': self.model.trainingwrapper['backbone'].parameters(), 'lr': lr},
            {'params': self.model.trainingwrapper['head'].parameters(), 'lr': lr * 10}
        ]
        return params


import torch
import torch.nn as nn
from typing import Callable
from functools import wraps
import torch.nn.functional as F
from torch import Tensor

__all__ = ['bce',
           'ce',
           'focal',
           'create_Lossfn',
           'list_lossfns',]

LOSS = {}

def register_loss(fn: Callable):
    key = fn.__name__
    if key in LOSS:
        raise ValueError(f"An entry is already registered under the name '{key}'.")
    LOSS[key] = fn
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    return wrapper

class FocalLoss(nn.Module):
    # Wraps focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, loss_fcn, gamma=1.5, alpha= 0.25):
        super().__init__()
        self.loss_fcn = loss_fcn  # must be nn.BCEWithLogitsLoss()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = loss_fcn.reduction
        self.loss_fcn.reduction = 'none'  # required to apply FL to each element

    def forward(self, pred, true):
        loss = self.loss_fcn(pred, true)
        # p_t = torch.exp(-loss)
        # loss *= self.alpha * (1.000001 - p_t) ** self.gamma  # non-zero power for gradient stability

        # TF implementation https://github.com/tensorflow/addons/blob/v0.7.1/tensorflow_addons/losses/focal_loss.py
        pred_prob = torch.sigmoid(pred)  # prob from logits
        p_t = true * pred_prob + (1 - true) * (1 - pred_prob)
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = (1.0 - p_t) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:  # 'none'
            return loss

class DistillKL(nn.Module):
    """Distilling the Knowledge in a Neural Network"""
    def __init__(self, T: float):
        super(DistillKL, self).__init__()
        self.T = T

    def forward(self, y_s: Tensor, y_t: Tensor):
        p_s = F.log_softmax(y_s/self.T, dim=1)
        p_t = F.softmax(y_t/self.T, dim=1)
        loss = F.kl_div(p_s, p_t, size_average=False) * (self.T**2) / y_s.shape[0]
        return loss

@register_loss
def bce():
    return nn.BCEWithLogitsLoss()
@register_loss
def ce(label_smooth: float = 0.):
    return nn.CrossEntropyLoss(label_smoothing=label_smooth)
@register_loss
def focal(gamma=1.5, alpha= 0.25):
    return FocalLoss(loss_fcn = nn.BCEWithLogitsLoss(), alpha=alpha, gamma=gamma)
def create_Lossfn(lossfn: str):
    lossfn = lossfn.strip()
    return LOSS[lossfn]

def list_lossfns():
    lossfns = [k for k, v in LOSS.items()]
    return sorted(lossfns)



from copy import deepcopy
import torch.nn as nn
from ...built.attention_based_pooler import atten_pool_replace
import torch
import torch.distributed as dist
import timm  
from torch.nn.init import normal_, constant_

class VisionWrapper:
    def __init__(self, model_cfgs: dict, logger = None, rank = -1):
        self.logger = logger
        self.model_cfgs = model_cfgs
        self.rank = rank

        self.kwargs = model_cfgs['kwargs']
        self.num_classes = model_cfgs['num_classes']
        self.pretrained = model_cfgs['pretrained']
        self.backbone_freeze = model_cfgs['backbone_freeze']
        self.bn_freeze = model_cfgs['bn_freeze']
        self.bn_freeze_affine = model_cfgs['bn_freeze_affine']

        model_cfgs_copy = deepcopy(model_cfgs)
        _, model_cfgs_copy['choice'] = model_cfgs['name'].split('-')

        self.model = self.create_model(**model_cfgs_copy)
        # pool layer
        if model_cfgs['attention_pool']:
            self.model = atten_pool_replace(self.model)

        del model_cfgs_copy

        if not self.pretrained: self.reset_parameters()

    def create_model(self, choice: str, num_classes: int = 1000, pretrained: bool = False, 
                    backbone_freeze: bool = False, bn_freeze: bool = False, 
                    bn_freeze_affine: bool = False, **kwargs):
        # Only rank 0 downloads the pre-trained weights
        if pretrained and self.rank == 0:
            _ = timm.create_model(
                choice,
                pretrained=True,
                num_classes=num_classes,
                **kwargs['kwargs']
            )

        if self.rank >= 0: 
            dist.barrier(device_ids=[self.rank])
        
        model = timm.create_model(
            choice,
            pretrained=pretrained,
            num_classes=num_classes,
            **kwargs['kwargs']
        )

        if backbone_freeze: self.freeze_backbone()
        if bn_freeze: self.freeze_bn(bn_freeze_affine)

        return model

    def load_weight(self, load_from_path: str | dict, ema: bool = False, device: torch.device = None):
        if isinstance(load_from_path, str):
            checkpoint = torch.load(load_from_path, map_location='cpu', weights_only=False)
            if ema:
                weights = checkpoint['ema'].float().state_dict()
            else:
                weights = checkpoint['model']
        elif isinstance(load_from_path, dict):
            weights = load_from_path['ema'].float().state_dict() if ema else load_from_path['model']
        else:
            raise TypeError(f"load_from_path must be str or dict, got {type(load_from_path)}")
            
        self.model.load_state_dict(weights)
        return self.model.to(device)

    def init_parameters(self, m: nn.Module):
        if isinstance(m, nn.Conv2d):
            normal_(m.weight, mean=0, std=0.02)
        elif isinstance(m, nn.BatchNorm2d):
            constant_(m.weight, 1)
            constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            normal_(m.weight, mean=0, std=0.02)
            constant_(m.bias, 0)

    def reset_parameters(self):
        self.model.apply(self.init_parameters)

    def freeze_backbone(self):
        # Get the classifier module
        classifier = self.model.get_classifier()
        
        # Freeze all parameters except the classifier
        for _, m in self.model.named_modules():
            if m is not classifier:  # Skip the classifier module
                for p in m.parameters(recurse=False):
                    p.requires_grad_(False)
                    
        if self.rank <= 0:  # Only print on main process
            print('backbone freeze')

    def freeze_bn(self, bn_freeze_affine: bool = False):
        for m in self.model.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()
                if bn_freeze_affine:
                    m.weight.requires_grad_(False)
                    m.bias.requires_grad_(False)


from .representation import FaceTrainingWrapper
from .classifier import VisionWrapper


def get_model(model_cfg, logger, rank):
    assert 'task' in model_cfg, 'Task is not specified'

    match model_cfg['task']:
        case 'face' | 'cbir': return FaceTrainingWrapper(model_cfg, logger)
        case 'classification': return VisionWrapper(model_cfg, logger, rank)


from copy import deepcopy
import torch.nn as nn
import math

def is_parallel(model):
    # Returns True if model is of type DP or DDP
    return type(model) in (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel)


def de_parallel(model):
    # De-parallelize a model: returns single-GPU model if model is of type DP or DDP
    return model.module if is_parallel(model) else model

class ModelEMA:
    """ Updated Exponential Moving Average (EMA) from https://github.com/rwightman/pytorch-image-models
    Keeps a moving average of everything in the model state_dict (parameters and buffers)
    For EMA details see https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    """

    def __init__(self, model, decay=0.9999, tau=2000, updates=0):
        # Create EMA
        self.ema = deepcopy(de_parallel(model)).eval()  # FP32 EMA
        self.updates = updates  # number of EMA updates
        self.decay = lambda x: decay * (1 - math.exp(-x / tau))  # decay exponential ramp (to help early epochs)
        for p in self.ema.parameters():
            p.requires_grad_(False)

    def update(self, model):
        # Update EMA parameters
        self.updates += 1
        d = self.decay(self.updates)

        msd = de_parallel(model).state_dict()  # model state_dict
        for k, v in self.ema.state_dict().items():
            if v.dtype.is_floating_point:  # true for FP16 and FP32
                v *= d
                v += (1 - d) * msd[k].detach()
        # assert v.dtype == msd[k].dtype == torch.float32, f'{k}: EMA {v.dtype} and model {msd[k].dtype} must be FP32'


import torch.nn as nn
from torch import Tensor
from typing import Callable
from ..engine.vision_engine import CenterProcessor

class Distiller:
    def __init__(self,
                 model_teacher: nn.Module,
                 model_student: nn.Module,
                 criterion_cls: Callable,
                 criterion_kl: Callable,
                 cls_weight: float = 0.5,
                 kl_weight: float = 0.5):

        self.model_teacher = model_teacher
        self.model_student = model_student
        self.criterion_cls = criterion_cls
        self.criterion_kl = criterion_kl
        self.cls_weight = cls_weight
        self.kl_weight = kl_weight

    def __call__(self, inputs: Tensor, label: Tensor) -> Tensor:
        # forward
        logit_s = self.model_student(inputs)
        logit_t = self.model_teacher(inputs)
        # compute loss
        loss_cls = self.criterion_cls(logit_s, label)
        loss_kl = self.criterion_kl(logit_s, logit_t)
        loss = self.cls_weight * loss_cls + self.kl_weight * loss_kl

        return loss

class DistillCenterProcessor(CenterProcessor):
    def __init__(self, cfgs: dict, rank: int, project: str, logger = None, opt = None):
        super().__init__(cfgs=cfgs['student'], rank=rank, project= project)

        # init teacher model
        # self.teacher = TorchVisionWrapper(cfgs['teacher'], logger = logger)
        # self.opt = opt































from typing import Union, List, Dict, Optional
from ..dataset.transforms import BaseClassWiseAugmenter

class ClassWiseAugmenter(BaseClassWiseAugmenter):
    def __init__(self, base_transforms: Dict, class_transforms_mapping: Optional[Dict[str, List[int]]], base: List[int]):
        if base is not None:
            assert isinstance(base, list), f'{base} is not a list of indices'
            base_transforms = [t for i, t in enumerate(base_transforms) if i in base]

        super().__init__(base_transforms=base_transforms, class_transforms_mapping=class_transforms_mapping)

    def __call__(self, image, label: Union[List, int], class_indices: List[int]):
        if self.class_transforms is None:
            return super().__call__(image=image, label=label, class_indices=class_indices)

        # softmax
        if isinstance(label, int):
            if class_indices[label] in self.class_transforms:
                return self.class_transforms[class_indices[label]](image)
            else: return self.base_transforms(image)
        # sigmoid
        elif isinstance(label, list): # multi-label
            # multi-label
            if len(label) == 1: # Customized specific class
                c = label[0]
                if class_indices[c] in self.class_transforms:
                    return self.class_transforms[class_indices[c]](image)

            # Generally common class
            return self.base_transforms(image)


from .timm_wrapper import TimmWrapper

class BackboneFactory:
    """Factory to produce backbone according the backbone_conf.yaml.
    
    Attributes:
        backbone_type: which backbone will produce.
        backbone_param:  params about model structure.
    """

    def __init__(self, backbone_config):
        # self.backbone_type = list(backbone_config['backbone'].keys())[0]
        # self.backbone_param = backbone_config['backbone'][self.backbone_type]
        for k, v in backbone_config.items(): self.backbone_type, self.backbone_param = k, v

    def get_backbone(self):

        if self.backbone_type.startswith('timm'):
            model = self.backbone_type.split('-')[1]
            return TimmWrapper(
                model_name=model,
                **self.backbone_param,
            )

        else:
            raise NotImplemented(f"{self.backbone_type} is not supported now !")


from functools import reduce
from pathlib import Path
import os
from datasets import load_dataset

def check_cfgs_common(cfgs):
    def find_normalize(augment_list):
        for augment in augment_list:
            if 'normalize' in augment:
                return augment['normalize']
        return None

    hyp_cfg = cfgs['hyp']
    data_cfg = cfgs['data']
    model_cfg = cfgs['model']

    # Check loss configuration
    assert reduce(lambda x, y: int(x) + int(y[0]), list(hyp_cfg['loss'].values())) == 1, \
        'Loss configuration error: Only one loss type should be enabled. Set either ce: true or bce: [true, ...] in hyp.loss'

    # Check optimizer
    assert hyp_cfg['optimizer'][0] in {'sgd', 'adam', 'sam'}, \
        'Invalid optimizer selection. Please choose from: sgd, adam, or sam'

    # Check scheduler and warm-up settings
    valid_schedulers = {'linear', 'cosine', 'linear_with_warm', 'cosine_with_warm'}
    assert hyp_cfg['scheduler'] in valid_schedulers, \
        'Invalid scheduler selection. Supported options: linear, cosine, linear_with_warm, cosine_with_warm'

    assert hyp_cfg['warm_ep'] >= 0 and isinstance(hyp_cfg['warm_ep'], int) and hyp_cfg['warm_ep'] < hyp_cfg['epochs'], \
        f'Invalid warm-up epochs: must be a non-negative integer less than total epochs ({hyp_cfg["epochs"]})'

    if hyp_cfg['warm_ep'] == 0:
        assert hyp_cfg['scheduler'] in {'linear', 'cosine'}, \
            'When warm-up is disabled (warm_ep: 0), only linear or cosine scheduler is supported'
    if hyp_cfg['warm_ep'] > 0:
        assert hyp_cfg['scheduler'] in {'linear_with_warm', 'cosine_with_warm'}, \
            'When using warm-up (warm_ep > 0), scheduler must be either linear_with_warm or cosine_with_warm'

    # Check normalization settings
    train_normalize = find_normalize(data_cfg['train']['augment'])
    val_normalize = find_normalize(data_cfg['val']['augment'])

    # Check backbone configuration
    if 'backbone' in model_cfg:
        backbone_cfg = next(iter(model_cfg['backbone'].items()))
        backbone_name, backbone_params = backbone_cfg
    else:
        backbone_name = model_cfg['name']
        backbone_params = {
            'pretrained': model_cfg.get('pretrained', False),
            'image_size': model_cfg.get('image_size')
        }

    # Verify model type is timm
    assert backbone_name.startswith('timm-'), \
        "Only timm models are supported. Model name must start with 'timm-'"

    # Check normalization requirements based on pretrained status
    is_pretrained = backbone_params.get('pretrained', False)
    if is_pretrained:
        if train_normalize is None or val_normalize is None:
            raise ValueError('Pretrained models require normalization in both training and validation augmentations')
        if train_normalize['mean'] != val_normalize['mean'] or train_normalize['std'] != val_normalize['std']:
            raise ValueError('Inconsistent normalization parameters: mean and std must be identical for training and validation')
    
    # Check image size for backbone
    assert 'image_size' in backbone_params, \
        f'Image size must be specified for {backbone_name}'
    assert backbone_params['image_size'] == model_cfg['image_size'], \
        f'Image size mismatch: {backbone_params["image_size"]} in backbone config vs {model_cfg["image_size"]} in model config'

def check_cfgs_face(cfgs):
    """
    Check configurations specific to face recognition tasks.
    
    Args:
        cfgs: Configuration dictionary
    """
    check_cfgs_common(cfgs=cfgs)

    model_cfg = cfgs['model']
    data_cfg = cfgs['data']

    # Check number of classes
    train_classes = [x for x in os.listdir(Path(data_cfg['root'])/'train') 
                    if not (x.startswith('.') or x.startswith('_'))]
    head_key = next(iter(model_cfg['head'].keys()))
    model_classes = model_cfg['head'][head_key]['num_class']
    
    assert model_classes == len(train_classes), \
        f'Model configuration error: Number of classes mismatch. Expected {len(train_classes)} from dataset, but got {model_classes} in model configuration'

    # Check face recognition specific configurations
    if cfgs['model']['task'] == 'face':
        pair_txt_path = data_cfg['val']['pair_txt']
        
        # Verify pair text file existence
        if not os.path.isfile(pair_txt_path):
            raise ValueError(f'Validation data error: Pair text file not found at {pair_txt_path}')

        # Validate pair list format
        from doraemon.engine.representation.eval_face import Evaluator
        try:
            with open(pair_txt_path) as f:
                pair_list = [line.strip() for line in f.readlines()]
            Evaluator.check_nps(pair_list)
        except Exception as e:
            raise ValueError(f'Pair list validation error: Invalid format in {pair_txt_path}. Details: {str(e)}')

def check_cfgs_cbir(cfgs):
    """
    Check configurations specific to CBIR (Content-Based Image Retrieval) tasks.
    
    Args:
        cfgs: Configuration dictionary
    """
    check_cfgs_common(cfgs=cfgs)

    model_cfg = cfgs['model']
    data_cfg = cfgs['data']

    # Determine data source type
    is_local = os.path.isdir(data_cfg['root'])

    # Check number of classes based on data source
    if is_local:
        train_classes = [x for x in os.listdir(Path(data_cfg['root'])/'train') 
                        if not (x.startswith('.') or x.startswith('_'))]
        num_classes = len(train_classes)
    else:
        try:
            dataset = load_dataset(data_cfg['root'], split='train')
            num_classes = len(set(dataset['label']))
        except Exception as e:
            raise ValueError(f"Dataset loading error: Unable to load HuggingFace dataset from {data_cfg['root']}. Details: {str(e)}")

    # Check model configuration
    head_key = next(iter(model_cfg['head'].keys()))
    model_classes = model_cfg['head'][head_key]['num_class']
    
    assert model_classes == num_classes, \
        f'Model configuration error: Number of classes mismatch. Expected {num_classes} from dataset, but got {model_classes} in model configuration'

def check_cfgs_classification(cfgs):
    """
    Check configurations specific to classification tasks.
    
    Args:
        cfgs: Configuration dictionary
    """
    check_cfgs_common(cfgs=cfgs)

    model_cfg = cfgs['model']
    data_cfg = cfgs['data']
    hyp_cfg = cfgs['hyp']

    # Determine data source type
    is_csv = data_cfg['root'].endswith('.csv')
    is_local = os.path.isdir(data_cfg['root'])

    # Check loss configuration based on data source
    if is_csv:
        if hyp_cfg['loss']['ce']:
            raise ValueError('Loss configuration error: Multi-label tasks (CSV format) require BCE loss. Please set ce: false in hyp.loss')
        if not hyp_cfg['loss']['bce'][0]:
            raise ValueError('Loss configuration error: Multi-label tasks (CSV format) require BCE loss. Please set bce: [true, ...] in hyp.loss')
    else:
        if not hyp_cfg['loss']['ce']:
            raise ValueError('Loss configuration error: Single-label tasks (folder structure/HuggingFace) require CE loss. Please set ce: true in hyp.loss')
        if hyp_cfg['loss']['bce'][0]:
            raise ValueError('Loss configuration error: Single-label tasks (folder structure/HuggingFace) do not support BCE loss. Please set bce: [false, ...] in hyp.loss')

    # Check num_classes
    if is_local:
        train_classes = [x for x in os.listdir(Path(data_cfg['root'])/'train') 
                        if not (x.startswith('.') or x.startswith('_'))]
        num_classes = len(train_classes)
    elif is_csv:
        import pandas as pd
        df = pd.read_csv(data_cfg['root'])
        class_columns = [col for col in df.columns if col not in ['image_path', 'train']]
        num_classes = len(class_columns)
        if hyp_cfg['loss']['bce'][0]:
            assert num_classes == len(hyp_cfg['loss']['bce'][1]), \
                f'Loss configuration error: Number of classes mismatch. Expected {len(hyp_cfg["loss"]["bce"][1])} from dataset, but got {num_classes} in model configuration'
    else:
        try:
            dataset = load_dataset(data_cfg['root'], split='train')
            num_classes = len(set(dataset['label']))
        except Exception as e:
            raise ValueError(f"Dataset loading error: Unable to load HuggingFace dataset from {data_cfg['root']}. Details: {str(e)}")

    assert model_cfg['num_classes'] == num_classes, \
        f'Model configuration error: Number of classes mismatch. Expected {num_classes} from dataset, but got {model_cfg["num_classes"]} in model configuration'

    # Check model configuration
    assert model_cfg['name'].split('-')[0] == 'timm', \
        'Model name error: Format should be [timm-ModelName] for timm models'

    if model_cfg['kwargs'] and model_cfg['pretrained']:
        for k in model_cfg['kwargs'].keys():
            if k not in {'dropout', 'attention_dropout', 'stochastic_depth_prob'}:
                raise KeyError('Model kwargs error: When using pretrained models, only [dropout, attention_dropout, stochastic_depth_prob] are allowed')

    # Check training strategies
    if hyp_cfg['strategy']['focal'][0]:
        assert hyp_cfg['loss']['bce'], \
            'Strategy configuration error: Focal loss requires BCE loss. Please enable BCE loss'
    
    if hyp_cfg['strategy']['ohem'][0]:
        assert not hyp_cfg['loss']['bce'][0], \
            'Strategy configuration error: OHEM is not compatible with BCE loss. Please disable BCE loss'

    # Check mixup configuration
    mix_ratio, mix_duration = hyp_cfg['strategy']['mixup']["ratio"], hyp_cfg['strategy']['mixup']["duration"]
    
    # Basic ratio check
    assert 0 <= mix_ratio <= 1, 'Mixup configuration error: ratio must be in [0,1]'
    
    # Only check duration when mixup is enabled
    if mix_ratio > 0:
        assert 0 < mix_duration <= hyp_cfg['epochs'], \
            f'Mixup configuration error: when mixup is enabled (ratio > 0), duration must be in (0,{hyp_cfg["epochs"]}]'
    
    hyp_cfg['strategy']['mixup'] = [mix_ratio, mix_duration]

def check(task, cfgs):
    if task == 'face': check_cfgs_face(cfgs)
    elif task == 'cbir': check_cfgs_cbir(cfgs)
    elif task == 'classification': check_cfgs_classification(cfgs)
    else: raise ValueError(f'{task} is not supported')


# based on:
# https://github.com/xiaoboCASIA/SV-X-Softmax/blob/master/fc_layers.py

import math
import torch
import torch.nn.functional as F
from torch.nn import Module, Parameter

class MV_Softmax(Module):
    """Implementation for "Mis-classified Vector Guided Softmax Loss for Face Recognition"
    """
    def __init__(self, feat_dim, num_class, is_am, margin=0.35, mv_weight=1.12, scale=32):
        super(MV_Softmax, self).__init__()
        self.weight = Parameter(torch.Tensor(feat_dim, num_class))
        self.weight.data.uniform_(-1, 1).renorm_(2, 1, 1e-5).mul_(1e5)
        self.margin = margin
        self.mv_weight = mv_weight
        self.scale = scale
        self.is_am = is_am
        self.cos_m = math.cos(margin)
        self.sin_m = math.sin(margin)
        self.threshold = math.cos(math.pi - margin)
        self.mm = self.sin_m * margin

    def forward(self, x, label):
        kernel_norm = F.normalize(self.weight, dim=0)
        x = F.normalize(x)
        cos_theta = torch.mm(x, kernel_norm)
        batch_size = label.size(0)
        gt = cos_theta[torch.arange(0, batch_size), label].view(-1, 1) 
        if self.is_am:  # AM
            mask = cos_theta > gt - self.margin
            final_gt = torch.where(gt > self.margin, gt - self.margin, gt)
        else:  # arcface
            sin_theta = torch.sqrt(1.0 - torch.pow(gt, 2))
            cos_theta_m = gt * self.cos_m - sin_theta * self.sin_m 
            mask = cos_theta > cos_theta_m
            final_gt = torch.where(gt > 0.0, cos_theta_m, gt)
        # process hard example.
        hard_example = cos_theta[mask]
        cos_theta[mask] = self.mv_weight * hard_example + self.mv_weight - 1.0
        cos_theta.scatter_(1, label.data.view(-1, 1), final_gt)
        cos_theta *= self.scale
        return cos_theta



import math
import torch
import torch.nn.functional as F
from torch.nn import Module, Parameter

class ArcFace(Module):
    """Implementation for "ArcFace: Additive Angular Margin Loss for Deep Face Recognition"
    """
    def __init__(self, feat_dim, num_class, margin_arc=0.35, margin_am=0.0, scale=32):
        super(ArcFace, self).__init__()
        self.weight = Parameter(torch.Tensor(feat_dim, num_class))
        self.weight.data.uniform_(-1, 1).renorm_(2, 1, 1e-5).mul_(1e5)
        self.margin_arc = margin_arc
        self.margin_am = margin_am
        self.scale = scale
        self.cos_margin = math.cos(margin_arc)
        self.sin_margin = math.sin(margin_arc)
        self.min_cos_theta = math.cos(math.pi - margin_arc)

    def forward(self, feats, labels):
        kernel_norm = F.normalize(self.weight, dim=0)
        feats = F.normalize(feats)
        cos_theta = torch.mm(feats, kernel_norm) 
        cos_theta = cos_theta.clamp(-1, 1)
        sin_theta = torch.sqrt(1.0 - torch.pow(cos_theta, 2))
        cos_theta_m = cos_theta * self.cos_margin - sin_theta * self.sin_margin
        # 0 <= theta + m <= pi, ==> -m <= theta <= pi-m
        # because 0<=theta<=pi, so, we just have to keep theta <= pi-m, that is cos_theta >= cos(pi-m)
        cos_theta_m = torch.where(cos_theta > self.min_cos_theta, cos_theta_m, cos_theta-self.margin_am)
        index = torch.zeros_like(cos_theta)
        index.scatter_(1, labels.data.view(-1, 1), 1)
        index = index.byte().bool()
        output = cos_theta * 1.0
        output[index] = cos_theta_m[index]
        output *= self.scale
        return output



from torch.optim import SGD, Adam
from typing import Callable, List, Dict, Optional, Iterator, Union
from functools import wraps
from abc import ABCMeta, abstractmethod
from torch.nn import Module
import torch
from torch.nn.modules.batchnorm import _BatchNorm

__all__ = ['sgd',
           'adam',
           'sam',
           'BaseSeperateLayer',
           'create_Optimizer',
           'list_optimizers']

OPTIMIZER = {}

def register_optimizer(fn: Callable):
    key = fn.__name__
    if key in OPTIMIZER:
        raise ValueError(f"An entry is already registered under the name '{key}'.")
    OPTIMIZER[key] = fn
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    return wrapper

class SAM(torch.optim.Optimizer):
    """
    https://arxiv.org/abs/2010.01412 Sharpness-Aware Minimization for Efficiently Improving Generalization
    """
    def __init__(self, params, base_optimizer, rho=0.05, adaptive=True, lr: float = 0.01, momentum: float = 0.9, weight_decay: float = 5e-4, **kwargs):
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"

        defaults = dict(rho=rho, adaptive=adaptive, lr=lr, momentum=momentum, weight_decay=weight_decay, **kwargs)
        super(SAM, self).__init__(params, defaults)

        self.base_optimizer = base_optimizer(self.param_groups, **kwargs)
        self.param_groups = self.base_optimizer.param_groups
        self.defaults.update(self.base_optimizer.defaults)

    @torch.no_grad()
    def first_step(self, zero_grad=False):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)

            for p in group["params"]:
                if p.grad is None: continue
                self.state[p]["old_p"] = p.data.clone()
                e_w = (torch.pow(p, 2) if group["adaptive"] else 1.0) * p.grad * scale.to(p)
                p.add_(e_w)  # climb to the local maximum "w + e(w)"

        if zero_grad: self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad=False):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None: continue
                p.data = self.state[p]["old_p"]  # get back to "w" from "w + e(w)"

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad: self.zero_grad()

    @torch.no_grad()
    def step(self, closure=None):
        assert closure is not None, "Sharpness Aware Minimization requires closure, but it was not provided"
        closure = torch.enable_grad()(closure)  # the closure should do a full forward-backward pass

        self.first_step(zero_grad=True)
        closure()
        self.second_step()

    def _grad_norm(self):
        shared_device = self.param_groups[0]["params"][0].device  # put everything on the same device, in case of model parallelism
        norm = torch.norm(
                    torch.stack([
                        ((torch.abs(p) if group["adaptive"] else 1.0) * p.grad).norm(p=2).to(shared_device)
                        for group in self.param_groups for p in group["params"]
                        if p.grad is not None
                    ]),
                    p=2
               )
        return norm

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self.base_optimizer.param_groups = self.param_groups

    def disable_running_stats(self, model):
        def _disable(module):
            if isinstance(module, _BatchNorm):
                module.backup_momentum = module.momentum
                module.momentum = 0

        model.apply(_disable)

    def enable_running_stats(self, model):
        def _enable(module):
            if isinstance(module, _BatchNorm) and hasattr(module, "backup_momentum"):
                module.momentum = module.backup_momentum

        model.apply(_enable)

class BaseSeperateLayer(metaclass=ABCMeta):
    """
        用于对model的多个层分别设置具体的学习率
    """
    def __init__(self, model: Module) -> None:
        self.model = model

    @abstractmethod
    def create_ParamSequence(self, alpha: Optional[float], lr: float) -> Union[Iterator, List[Dict]]:
        pass

@register_optimizer
def sgd(*args, **kwargs):
    return SGD(*args, **kwargs)

@register_optimizer
def adam(*args, **kwargs):
    return Adam(*args, **kwargs)

@register_optimizer
def sam(base_optimizer = SGD, *args, **kwargs):
    return SAM(base_optimizer=base_optimizer, *args, **kwargs)

def create_Optimizer(optimizer: str, lr: float, weight_decay, momentum, params):
    # return partial(OPTIMIZER[optimizer], lr = lr, weight_decay = weight_decay, momentum = momentum)
    return OPTIMIZER[optimizer](params = params, lr = lr, weight_decay = weight_decay, momentum = momentum)

def list_optimizers():
    optimizers = [k for k, v in OPTIMIZER.items()]
    return sorted(optimizers)


import yaml
from .arcface import ArcFace
from .circleloss import CircleLoss
from .mv_softmax import MV_Softmax
from .magface import MagFace

class HeadFactory:
    """Factory to produce head according to the head_conf.yaml
    
    Attributes:
        head_type(str): which head will be produce.
        head_param(dict): parsed params and it's value.
    """
    def __init__(self, head_config):
        for k, v in head_config.items(): self.head_type, self.head_param = k, v

    def get_head(self):
        if self.head_type == 'arcface':
            feat_dim = self.head_param['feat_dim'] # dimension of the output features, e.g. 512 
            num_class = self.head_param['num_class'] # number of classes in the training set.
            margin_arc = self.head_param['margin_arc'] # cos(theta + margin_arc).
            margin_am = self.head_param['margin_am'] # cos_theta - margin_am.
            scale = self.head_param['scale'] # the scaling factor for cosine values.
            head = ArcFace(feat_dim, num_class, margin_arc, margin_am, scale)

        elif self.head_type == 'magface':
            feat_dim = self.head_param['feat_dim'] # dimension of the output features, e.g. 512
            num_class = self.head_param['num_class'] # number of classes in the training set.
            margin_am = self.head_param['margin_am'] # cos_theta - margin_am.
            scale = self.head_param['scale'] # the scaling factor for cosine values.
            l_a = self.head_param['l_a']
            u_a = self.head_param['u_a']
            l_margin = self.head_param['l_margin']
            u_margin = self.head_param['u_margin']
            lamda = self.head_param['lamda']
            head = MagFace(feat_dim, num_class, margin_am, scale, l_a, u_a, l_margin, u_margin, lamda)

        elif self.head_type == 'circleloss':
            feat_dim = self.head_param['feat_dim'] # dimension of the output features, e.g. 512 
            num_class = self.head_param['num_class'] # number of classes in the training set.
            margin = self.head_param['margin'] # O_p = 1 + margin, O_n = -margin.
            gamma = self.head_param['gamma'] # the scale facetor.
            head = CircleLoss(feat_dim, num_class, margin, gamma)

        elif self.head_type == 'mv-softmax':
            feat_dim = self.head_param['feat_dim'] # dimension of the output features, e.g. 512 
            num_class = self.head_param['num_class'] # number of classes in the training set.
            is_am = self.head_param['is_am'] # am-softmax for positive samples.
            margin = self.head_param['margin'] # margin for positive samples.
            mv_weight = self.head_param['mv_weight'] # weight for hard negtive samples.
            scale = self.head_param['scale'] # the scaling factor for cosine values.
            head = MV_Softmax(feat_dim, num_class, is_am, margin, mv_weight, scale)

        else:
            raise NotImplemented("only arcface, magface, circleloss and mv-softmax are supported now !")
        return head



import torch
import torch.nn as nn
# from torchvision.models import MobileNetV3, ResNet, ConvNeXt, EfficientNet, SwinTransformer


class AttentionPooling(nn.Module):
    """
    Augmenting Convolutional networks with attention-based aggregation: https://arxiv.org/abs/2112.13692
    """
    def __init__(self, in_dim):
        super().__init__()
        self.cls_vec = nn.Parameter(torch.randn(in_dim))
        self.fc = nn.Linear(in_dim, in_dim)
        self.softmax = nn.Softmax(-1)

    def forward(self, x):
        # x.view: B,C,H,W -> BxHxW,C
        weights = torch.matmul(x.reshape(-1, x.shape[1]), self.cls_vec)
        # weights.view: BxHxW,C -> B, HxW
        weights = self.softmax(weights.reshape(x.shape[0], -1))
        # x.view: B,C,H,W -> B,C,HxW
        # (B,C,HxW) @ (B,HxW,1) -> B,C,1
        x = torch.bmm(x.reshape(x.shape[0], x.shape[1], -1), weights.unsqueeze(-1)).squeeze(-1)
        x = x + self.cls_vec
        x = self.fc(x)
        x = x + self.cls_vec
        return x

def atten_pool_replace(model: nn.Module):
    #--------------------------------Custom Pooling-----------------------------------#
    # if type(model) is MobileNetV3:
    #     model.avgpool = AttentionPooling(in_dim=model.classifier[0].in_features)
    # elif type(model) is ResNet:
    #     model.avgpool = AttentionPooling(in_dim=model.fc.in_features)
    # elif type(model) is ConvNeXt:
    #     model.avgpool = AttentionPooling(in_dim=model.classifier[-1].in_features)
    # elif type(model) is EfficientNet:
    #     model.avgpool = AttentionPooling(in_dim=model.classifier[-1].in_features)
    # elif type(model) is SwinTransformer:
    #     model.avgpool = AttentionPooling(in_dim=model.head.in_features)
    # else:
    #     raise KeyError(f'{type(model)} not support attention-based pool')

    return model
    #-----------------------------------------------------------------------------#


from .backbone.backbone_def import BackboneFactory
from .head.head_def import HeadFactory
import torch.nn as nn
from torch.nn.init import normal_, constant_
import torch
import torch.nn.functional as F
import os
import numpy as np

class FaceTrainingWrapper:
    def __init__(self, model_cfg, logger = None):
        self.model = FaceTrainingModel(model_cfg)
        self.logger = logger

    def init_parameters(self, m: nn.Module):
        if isinstance(m, nn.Conv2d):
            normal_(m.weight, mean=0, std=0.02)
        elif isinstance(m, nn.BatchNorm2d):
            constant_(m.weight, 1)
            constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            normal_(m.weight, mean=0, std=0.02)
            constant_(m.bias, 0)

    def reset_parameters(self):
        self.model.apply(self.init_parameters)

class FaceTrainingModel(nn.Module):
    """Define a traditional faceX model which contains a backbone and a head.

    Attributes:
        backbone: the backbone of faceX model.
        head: the head of faceX model.
    """

    def __init__(self, model_cfg):
        """Init faceX model by backbone factorcy and head factory.

        Args:
            backbone_factory: produce a backbone according to config files.
            head_factory: produce a head according to config files.
        """
        super().__init__()
        backbone = BackboneFactory(model_cfg['backbone']).get_backbone()
        head = HeadFactory(model_cfg['head']).get_head()
        self.trainingwrapper = nn.ModuleDict({
            'backbone': backbone,
            'head': head
        })

    def forward(self, data, label):
        feat = self.trainingwrapper['backbone'](data)
        pred = self.trainingwrapper['head'](feat, label)
        return pred

class FaceModelLoader:
    def __init__(self, model_cfg: dict):
        self.model = BackboneFactory(model_cfg['backbone']).get_backbone()

    def load_weight_default(self, model_path):
        """The default method to load a model.

        Args:
            model_path:: the path of the weight file.

        Returns:
            model: initialized model.
        """
        self.model.load_state_dict(torch.load(model_path, weights_only=False)['state_dict'], strict=True)

        return self.model

    def load_weight(self, model_path, ema: bool = False):
        """The custom method to load a model, from a model having feature extractor and head.

        Args:
            model_path: the path of the weight file.

        Returns:
            model: initialized model.
        """
        pretrained_dict = torch.load(model_path, weights_only=False)['ema'] if ema else torch.load(model_path, weights_only=False)['state_dict']

        self.model.load_state_dict(pretrained_dict, strict=True)

        return self.model

class FeatureExtractor:

    def __init__(self, model):
        self.model = model

    def extract_face(self, dataloader, device) -> dict:
        """Extract and return features.

        Args:
            model: initialized model.
            dataloader: load data to be extracted.

        Returns:
            image_name2feature: key is the name of image, value is feature of image.
        """
        model = self.model
        model.eval()
        model.to(device)

        image_name2feature = {}
        with torch.no_grad():
            for batch_idx, (_, tensors, file_realpaths) in enumerate(dataloader):
                tensors = tensors.to(device)
                features = model(tensors)
                features = F.normalize(features)
                features = features.cpu().numpy()
                for realpath, feature in zip(file_realpaths, features):
                    filename = os.path.join(os.path.basename(os.path.dirname(realpath)), os.path.basename(realpath))
                    image_name2feature[filename] = feature

        return image_name2feature
    
    def extract_cbir(self, dataloader, device) -> np.ndarray:
        """Extract and return features.

        Args:
            model: initialized model.
            dataloader: load data to be extracted.

        Returns:
            features: feature of image.
        """
        model = self.model
        model.eval()
        model.to(device)

        features = []
        with torch.no_grad():
            for batch_idx, tensors in enumerate(dataloader):
                tensors = tensors.to(device)
                feature = model(tensors)
                feature = F.normalize(feature)
                feature = feature.cpu().numpy()

                features.append(feature)

        return np.concatenate(features, axis=0)


from typing import Tuple, Callable, Optional, List
import numpy as np
import torch
import torch.nn as nn

from timm.models import VisionTransformer, \
    SwinTransformer, \
    ResNet, \
    MobileNetV3, \
    ConvNeXt, \
    SwinTransformerV2, \
    SENet, \
    EfficientNet

from PIL.Image import Image as ImageType
from torchvision.transforms import Compose
from ..dataset.transforms import SPATIAL_TRANSFORMS
from ..dataset.transforms import PadIfNeed, Reverse_PadIfNeed, ResizeAndPadding2Square, ReverseResizeAndPadding2Square
import cv2

from pytorch_grad_cam import GradCAM, \
    ScoreCAM, \
    GradCAMPlusPlus, \
    AblationCAM, \
    XGradCAM, \
    EigenCAM, \
    EigenGradCAM, \
    LayerCAM, \
    FullGrad

from pytorch_grad_cam.utils.image import show_cam_on_image
from pytorch_grad_cam.ablation_layer import AblationLayerVit

# -----------------ClassifierOutputTarget is used to specify the targets, specifically which class the model is looking at---------------------- #
from pytorch_grad_cam.utils.model_targets import ClassifierOutputTarget
# ---------------------------------------------------------------------------------------------- #

class ClassActivationMaper:

    METHODS = \
        {"gradcam": GradCAM,
         "scorecam": ScoreCAM,
         "gradcam++": GradCAMPlusPlus,
         "ablationcam": AblationCAM,
         "xgradcam": XGradCAM,
         "eigencam": EigenCAM,
         "eigengradcam": EigenGradCAM,
         "layercam": LayerCAM,
         "fullgrad": FullGrad}

    def __init__(self, model: nn.Module, method: str, device, transforms):
        self.model = model
        self.device = device

        target_layers, reshape_transform = self._create_target_layers_and_transform(model)

        # cam
        if method not in self.METHODS:
            raise Exception(f"Method {method} not implemented")
        if method == "ablationcam":
            self.cam = ClassActivationMaper.METHODS[method](model=model,
                                       target_layers=target_layers,
                                       reshape_transform=reshape_transform,
                                       use_cuda = device != torch.device('cpu'),
                                       ablation_layer=AblationLayerVit())
        else:
            self.cam = ClassActivationMaper.METHODS[method](model=model,
                                       target_layers=target_layers,
                                       use_cuda = device != torch.device('cpu'),
                                       reshape_transform=reshape_transform)

        self.spatial_transforms, reversed_fun = ClassActivationMaper.pickup_spatial_transforms(transforms)

        if reversed_fun is not None:
            self.reverse_pad2square = reversed_fun

    def __call__(self,
                 image,
                 input_tensor: torch.Tensor,
                 dsize: Tuple[int, int],# w, h
                 targets: Optional[List[ClassifierOutputTarget]] = None) -> np.array:
        grayscale_cam = self.cam(input_tensor=input_tensor,
                                 targets=targets,
                                 eigen_smooth=False,
                                 aug_smooth=False)

        grayscale_cam = grayscale_cam[0, :]

        if not isinstance(image, ImageType):
            raise ValueError("Only images read by PIL.Image are allowed")

        image = self.spatial_transforms(image)
        image = np.array(image, dtype=np.float32)

        cam_image = show_cam_on_image(image / 255, grayscale_cam)

        if hasattr(self, 'reverse_pad2square'):
            if max(dsize) != max(cam_image.shape):
                if isinstance(self.reverse_pad2square, Reverse_PadIfNeed):
                    cam_image = cv2.resize(cam_image, (max(dsize), max(dsize)), cv2.INTER_LINEAR)
                elif isinstance(self.reverse_pad2square, ReverseResizeAndPadding2Square):
                    pass
                else:
                    raise ValueError(f"{type(self.reverse_pad2square)} not support reverse function")
                cam_image = self.reverse_pad2square(cam_image, dsize)
        return cam_image

    def _create_target_layers_and_transform(self, model: nn.Module) -> Tuple[list, Optional[Callable]]:

        if isinstance(model, (SwinTransformer, SwinTransformerV2)):
            return [model.norm], lambda tensor: torch.permute(tensor, dims=[0, 3, 1, 2])

        elif isinstance(model, VisionTransformer):
            # Get patch and image size information
            patch_size = model.patch_embed.patch_size
            if isinstance(patch_size, int):
                patch_size = (patch_size, patch_size)
            
            img_size = model.patch_embed.img_size
            if isinstance(img_size, int):
                img_size = (img_size, img_size)
            
            # Calculate feature map size
            feature_size = (img_size[0] // patch_size[0],
                           img_size[1] // patch_size[1])
            
            def reshape_transform(tensor):
                # Remove CLS token
                tensor = tensor[:, 1:, :]
                
                # Reshape to [batch_size, height, width, channels]
                B, _, C = tensor.shape
                H, W = feature_size
                tensor = tensor.reshape(B, H, W, C)
                
                # Convert to [batch_size, channels, height, width]
                tensor = tensor.permute(0, 3, 1, 2)
                return tensor
            
            return [model.blocks[-1].norm1], reshape_transform

        elif isinstance(model, MobileNetV3):
            return [model.blocks[-1][0].conv], None

        elif isinstance(model, (SENet, ResNet)):
            return [model.layer4[-1].conv3], None

        elif isinstance(model, ConvNeXt):
            return [model.norm_pre], None

        elif isinstance(model, EfficientNet):
            return [model.bn2], None

        else:
            raise KeyError(f'{type(model)} not support yet')

    @staticmethod
    def pickup_spatial_transforms(transforms: Compose):
        sequence = []
        reversed_fun = None
        for t in transforms.transforms:
            if type(t) in SPATIAL_TRANSFORMS:
                sequence.append(t)
            if type(t) is PadIfNeed:
                reversed_fun = Reverse_PadIfNeed(mode=t.mode)
            elif type(t) is ResizeAndPadding2Square:
                reversed_fun = ReverseResizeAndPadding2Square(size=t.size)

        return Compose(sequence), reversed_fun


from ...dataset.basedataset import CBIRDatasets
from ...dataset.transforms import create_AugTransforms
from ...dataset.dataprocessor import SmartDataProcessor
from ...models.representation.face_model import FeatureExtractor
from ...utils.logger import SmartLogger
import torch
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm
import faiss
from typing import Optional
from sklearn.metrics import roc_auc_score, ndcg_score

class CBIRMetrics:
    def __init__(self, 
                 cutoffs: list[int] = [1,10, 100]):
        _cutoffs = cutoffs.copy()
        self.cutoffs = _cutoffs
        self.metrics = {}

    def compute_mrr(self, 
                    preds: list[list[str]], 
                    labels: list[list[str]]):
        cutoffs = self.cutoffs.copy()
        mrrs = np.zeros(len(cutoffs))

        for pred, label in zip(preds, labels):
            jump = False
            for i, x in enumerate(pred, 1):
                if x in label:
                    for k, cutoff in enumerate(cutoffs):
                        if i <= cutoff:
                            mrrs[k] += 1 / i
                    jump = True
                if jump:
                    break
        mrrs /= len(preds)
        for i, cutoff in enumerate(cutoffs):
            mrr = mrrs[i]
            self.metrics[f"MRR@{cutoff}"] = mrr
        
    def compute_recall(self,
                       preds: list[list[str]],
                       labels: list[list[str]]):
        cutoffs = self.cutoffs.copy()
        recalls = np.zeros(len(cutoffs))
        for pred, label in zip(preds, labels):
            for k, cutoff in enumerate(cutoffs):
                recall = np.intersect1d(label, pred[:cutoff])
                recalls[k] += len(recall) / len(label)
        recalls /= len(preds)
        for i, cutoff in enumerate(cutoffs):
            recall = recalls[i]
            self.metrics[f"Recall@{cutoff}"] = recall
        
    def compute_precision(self,
                       preds: list[list[str]],
                       labels: list[list[str]]):
        cutoffs = self.cutoffs.copy()
        precisions = np.zeros(len(cutoffs))
        cutoffs = self.cutoffs.copy()
        for pred, label in zip(preds, labels):
            for k, cutoff in enumerate(cutoffs):
                precision = np.intersect1d(label, pred[:cutoff])
                precisions[k] += len(precision) / min(cutoff, len(label))
        precisions /= len(preds)
        for i, cutoff in enumerate(cutoffs):
            self.metrics[f"Precision@{cutoff}"] = precisions[i]

    def compute_auc(self,
                    preds: list[list[str]],
                    labels: list[list[str]],
                    preds_scores: list[list[float]]):
        pred_hard_encodings = self.encode_pred2hard(preds=preds, labels=labels)
        
        pred_hard_encodings1d = np.asarray(pred_hard_encodings).flatten() 
        preds_scores1d = preds_scores.flatten()
        auc = roc_auc_score(pred_hard_encodings1d, preds_scores1d)
        
        self.metrics[f'AUC@{self.cutoffs[-1]}'] = auc
    
    def compute_ndcg(self,
                     preds: list[list[str]],
                     labels: list[list[str]],
                     preds_scores: list[list[float]]):
        cutoffs = self.cutoffs.copy()
        pred_hard_encodings = self.encode_pred2hard(preds=preds, labels=labels)
        for _, cutoff in enumerate(cutoffs):
            nDCG = ndcg_score(pred_hard_encodings, preds_scores, k=cutoff)
            self.metrics[f"nDCG@{cutoff}"] = nDCG

    def encode_pred2hard(self,
                         preds: list[list[str]],
                         labels: list[list[str]]) -> list[list[int]]:
        pred_hard_encodings = []
        for pred, label in zip(preds, labels):
            pred_hard_encoding = np.isin(pred, label).astype(int).tolist()
            pred_hard_encodings.append(pred_hard_encoding)
        
        return pred_hard_encodings

    def reset(self):
        self.metrics.clear()


def index(extractor: FeatureExtractor, 
        gallery_dataloader: DataLoader, 
        device: torch.device,
        logger: SmartLogger,
        index_factory: str = "Flat", 
        # need memmap
        memmap_feat_dim: Optional[int] = None,
        memmap_dtype: torch.dtype = torch.float16,
        memmap_save_path: Optional[str] = None, 
        memmap_load_embedding: bool = False,
        ):

    """
    1. Encode the entire corpus into dense embeddings; 
    2. Create faiss index; 
    3. Optionally save embeddings.
    """

    if memmap_load_embedding:
        gallery_embeddings = np.memmap(
            memmap_save_path,
            mode="r",
            dtype=memmap_dtype
        ).reshape(-1, memmap_feat_dim)
    
    else:
        gallery_embeddings = extractor.extract_cbir(gallery_dataloader, device)
        
        if memmap_save_path is not None:
            logger.console(f"saving embeddings at {memmap_save_path}...")
            memmap = np.memmap(
                memmap_save_path,
                shape=gallery_embeddings.shape,
                mode="w+",
                dtype=gallery_embeddings.dtype
            )

            length = gallery_embeddings.shape[0]
            # add in batch
            save_batch_size = 10000
            if length > save_batch_size:
                for i in tqdm(range(0, length, save_batch_size), leave=False, desc="Saving Embeddings"):
                    j = min(i + save_batch_size, length)
                    memmap[i: j] = gallery_embeddings[i: j]
            else:
                memmap[:] = gallery_embeddings
    
    dim = gallery_embeddings.shape[-1]
    # create faiss index

    faiss_index = faiss.index_factory(dim, index_factory, faiss.METRIC_INNER_PRODUCT)
    logger.console(f"Creating CPU FAISS index with dimension {dim}...")

    # if device.type == 'cuda':
    #     # co = faiss.GpuClonerOptions()
    #     co = faiss.GpuMultipleClonerOptions()
    #     co.useFloat16 = True
    #     # faiss_index = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(), 0, faiss_index, co)
    #     faiss_index = faiss.index_cpu_to_all_gpus(faiss_index, co)

    # NOTE: faiss only accepts float32
    logger.console("Adding embeddings...")
    gallery_embeddings = gallery_embeddings.astype(np.float32)
    faiss_index.train(gallery_embeddings)
    faiss_index.add(gallery_embeddings)
    return faiss_index

def search(extractor: FeatureExtractor, 
        query_dataloader: DataLoader, 
        faiss_index: faiss.Index, 
        device: torch.device,
        logger: SmartLogger,
        k:int = 100, 
        batch_size: int = 256,
        ):
    """
    1. Encode queries into dense embeddings;
    2. Search through faiss index
    """
    query_embeddings = extractor.extract_cbir(query_dataloader, device)
    query_size = query_embeddings.shape[0] 
    
    all_scores = []
    all_indices = []
    
    logger.console('Searching ...')
    for i in range(0, query_size, batch_size):
        j = min(i + batch_size, query_size)
        query_embedding = query_embeddings[i: j]
        score, indice = faiss_index.search(query_embedding.astype(np.float32), k=k)
        all_scores.append(score)
        all_indices.append(indice)
    
    all_scores = np.concatenate(all_scores, axis=0)
    all_indices = np.concatenate(all_indices, axis=0)

    return all_scores, all_indices

def compute_metrics(preds, 
                    preds_scores, 
                    labels, 
                    metrics = ['mrr', 'precision', 'recall', 'auc', 'ndcg'],
                    cutoffs=[1, 3, 10]):

    metrics_engine = CBIRMetrics(cutoffs=cutoffs)

    for m in metrics:
        if m == 'mrr':
            metrics_engine.compute_mrr(preds=preds, labels=labels)
        elif m == 'precision':
            metrics_engine.compute_precision(preds=preds, labels=labels)
        elif m == 'recall':
            metrics_engine.compute_recall(preds=preds, labels=labels)
        elif m == 'auc':
            metrics_engine.compute_auc(preds=preds, labels=labels, preds_scores=preds_scores)
        elif m == 'ndcg':
            metrics_engine.compute_ndcg(preds=preds, labels=labels, preds_scores=preds_scores)
        else:
            raise ValueError(f'{m} is not supported')

    return metrics_engine.metrics

def valuate(model,
            data_cfg: dict,
            device: torch.device,
            logger: SmartLogger,
            vis: bool = False):
    """
    Arguments:
        vis(bool): for cbir visualization
    """
    
    query_dataset, gallery_dataset = CBIRDatasets.build(root=data_cfg['root'], 
                                                        transforms=create_AugTransforms(data_cfg['val']['augment']))
    
    query_dataloader = SmartDataProcessor.set_dataloader(query_dataset, 
                                                         bs=data_cfg['val']['bs'], 
                                                         nw=data_cfg['nw'], 
                                                         shuffle=False) # must be False, otherwise metrics are computed wrong

    gallery_dataloader = SmartDataProcessor.set_dataloader(gallery_dataset, 
                                                           bs=data_cfg['val']['bs'], 
                                                           nw=data_cfg['nw'], 
                                                           shuffle=False) # must be False, otherwise metrics are computed wrong
                                        
    feature_extractor = FeatureExtractor(model)

    faiss_index = index(
        extractor=feature_extractor, 
        gallery_dataloader=gallery_dataloader,
        device=device,
        logger=logger
        )

    cutoffs = data_cfg['val']['metrics']['cutoffs']
    scores, indices = search(
        extractor=feature_extractor,
        query_dataloader=query_dataloader,
        faiss_index=faiss_index, 
        device=device,
        logger=logger,
        k = cutoffs[-1],
        batch_size=data_cfg['val']['bs'],
        )

    retrieval_results = []
    for indice in indices:
        # filter invalid indices
        indice = indice[indice != -1].tolist()
        retrieval_results.append(gallery_dataset.gallery[indice]['gallery'])

    ground_truths = []
    for pos in query_dataset.data['pos']:
        ground_truths.append(pos)

    metrics = compute_metrics(retrieval_results, 
                              scores, 
                              ground_truths, 
                              metrics=data_cfg['val']['metrics']['metrics'],
                              cutoffs=cutoffs)

    for k, v in metrics.items():
        metrics[k] = float(v)
    
    if vis:
        return metrics, retrieval_results, scores, ground_truths, query_dataset.data['query'], query_dataset, gallery_dataset

    return metrics




import torch
import torch.nn as nn
from torch.amp import autocast
from tqdm import tqdm
from typing import Callable, Optional, Union, List
from torchmetrics import Precision, Recall, F1Score
from torch import Tensor
import itertools
import matplotlib.pyplot as plt
import numpy as np
from typing import Sequence
from prettytable import PrettyTable


__all__ = ['valuate']

class ConfusedMatrix:
    def __init__(self, nc: int):
        self.nc = nc
        self.mat = None

    def update(self, gt: Tensor, pred: Tensor):
        if self.mat is None: self.mat = torch.zeros((self.nc, self.nc), dtype=torch.int64, device = gt.device)

        idx = gt * self.nc + pred
        self.mat += torch.bincount(idx, minlength=self.nc).reshape(self.nc, self.nc)

    def save_conm(self, cm: np.ndarray, classes: Sequence, save_path: str, cmap=plt.cm.cool):
        """
        - cm : 计算出的混淆矩阵的值
        - classes : 混淆矩阵中每一行每一列对应的列
        - normalize : True:显示百分比, False:显示个数
        """
        ax = plt.gca()
        ax.tick_params(axis="x", top=True, labeltop=True, bottom=False, labelbottom=False)
        plt.imshow(cm, interpolation='nearest', cmap=cmap)
        plt.colorbar()
        tick_marks = [x for x in range(len(classes))]
        plt.xticks(tick_marks, classes, rotation=0, fontsize=10)
        plt.yticks(tick_marks, classes, fontsize=10)
        fmt = '.2f'
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            plt.text(j, i, format(cm[i, j], fmt),
                     horizontalalignment="center",
                     color="black")
        plt.tight_layout()
        plt.ylabel('GT', fontsize=12)
        plt.xlabel('Predict', fontsize=12)
        ax.xaxis.set_label_position('top')
        plt.gcf().subplots_adjust(top=0.9)
        plt.savefig(save_path)

def valuate(model: nn.Module, dataloader, device: torch.device, pbar, 
            is_training: bool = False, lossfn: Optional[Callable] = None, 
            logger = None, thresh: Union[float, List[float]] = 0, 
            top_k: int = 5, conm_path: str = None):
    """
    Evaluate model performance
    
    Args:
        ...
        thresh: float or List[float], threshold for each class in multi-label classification
               - float: same threshold for all classes
               - List[float]: specific threshold for each class
        ...
    """
    is_single_label = isinstance(thresh, (int, float)) and thresh == 0

    # Check threshold type
    if isinstance(thresh, (list, tuple, np.ndarray)):
        # Multi-threshold case: each class uses a different threshold
        assert len(thresh) == len(dataloader.dataset.id2label), \
            f'Number of thresholds ({len(thresh)}) must match number of classes ({len(dataloader.dataset.id2label)})'
        thresh = torch.tensor(thresh, device=device)
        # Verify that all thresholds are within the valid range
        assert (thresh > 0).all() and (thresh < 1).all(), \
            'For multi-label (BCE), all thresholds should be in (0, 1)'
    elif isinstance(thresh, (int, float)):
        if is_single_label:
            # Single-label classification (Softmax) case
            pass
        else:
            # Multi-label classification (BCE), use the same threshold
            assert 0 < thresh < 1, 'For multi-label (BCE), threshold should be in (0, 1)'
            thresh = torch.full((len(dataloader.dataset.id2label),), 
                              thresh, 
                              device=device)
    else:
        raise ValueError(f'Unsupported threshold type: {type(thresh)}. '
                        f'Expected float or list/tuple/ndarray of floats.')

    # eval mode
    model.eval()

    n = len(dataloader)  # number of batches
    action = 'validating'
    desc = f'{pbar.desc[:-36]}{action:>36}' if pbar else f'{action}'
    bar = tqdm(dataloader, desc, n, not is_training, bar_format='{l_bar}{bar:10}{r_bar}', position=0)
    pred, targets, loss = [], [], 0
    
    with torch.no_grad():
        with autocast('cuda', enabled=(device != torch.device('cpu'))):
            for images, labels in bar:
                images, labels = images.to(device, non_blocking=True), labels.to(device)
                y = model(images)
                if is_single_label:
                    pred.append(y.argsort(1, descending=True)[:, :top_k])
                    targets.append(labels)
                else:
                    # Get prediction probabilities using sigmoid
                    pred_prob = y.sigmoid()
                    # Predict using threshold for each class
                    pred.append(pred_prob >= thresh)
                    # Convert to hard labels
                    hard_labels = (labels >= 0.5).float()
                    targets.append(hard_labels)
                if lossfn:
                    loss += lossfn(y, labels)

    loss /= n
    pred, targets = torch.cat(pred), torch.cat(targets)
    
    if not is_training and is_single_label and len(dataloader.dataset.id2label) <= 10:
        conm = ConfusedMatrix(len(dataloader.dataset.id2label))
        conm.update(targets, pred[:, 0])
        conm.save_conm(conm.mat.detach().cpu().numpy(), dataloader.dataset.id2label, conm_path if conm_path is not None else 'conm.png')

    if is_single_label:
        correct = (targets[:, None] == pred).float()
        acc = torch.stack((correct[:, 0], correct.max(1).values), dim=1)  # (top1, top5) accuracy
        top1, top5 = acc.mean(0).tolist()
        
        if not is_training:
            table = PrettyTable(['Class', 'Samples', 'Top1', f'Top{top_k}'])
            
            for i, c in dataloader.dataset.id2label.items():
                acc_i = acc[targets == i]
                top1i, top5i = acc_i.mean(0).tolist()
                table.add_row([c, acc_i.shape[0], f'{top1i:.3f}', f'{top5i:.3f}'])
            
            table.add_row(['MEAN', acc.shape[0], f'{top1:.3f}', f'{top5:.3f}'])
            
            logger.console('\n' + str(table))
        else:
            table = PrettyTable(['Class', 'Samples', 'Top1', f'Top{top_k}'])
            for i, c in dataloader.dataset.id2label.items():
                acc_i = acc[targets == i]
                top1i, top5i = acc_i.mean(0).tolist()
                table.add_row([c, acc_i.shape[0], f'{top1i:.3f}', f'{top5i:.3f}'])
            table.add_row(['MEAN', acc.shape[0], f'{top1:.3f}', f'{top5:.3f}'])
            logger.log('\n' + str(table))
    else:
        num_classes = len(dataloader.dataset.id2label)
        # Compute precision, recall, and F1-score for each class
        precisioner = Precision(task='multilabel', threshold=0.5, num_labels=num_classes, average=None).to(device)
        recaller = Recall(task='multilabel', threshold=0.5, num_labels=num_classes, average=None).to(device)
        f1scorer = F1Score(task='multilabel', threshold=0.5, num_labels=num_classes, average=None).to(device)

        # Compute precision, recall, and F1-score for each class
        precision = precisioner(pred.float(), targets)
        recall = recaller(pred.float(), targets)
        f1score = f1scorer(pred.float(), targets)

        cls_numbers = targets.sum(0).int().tolist()
        
        if is_training:
            table = PrettyTable(['Class', 'Samples', 'Precision', 'Recall', 'F1-Score', 'Threshold'])
            for i, c in dataloader.dataset.id2label.items():
                table.add_row([
                    c, 
                    cls_numbers[i], 
                    f'{precision[i].item():.3f}',
                    f'{recall[i].item():.3f}',
                    f'{f1score[i].item():.3f}',
                    f'{thresh[i].item():.3f}' if isinstance(thresh, torch.Tensor) else f'{thresh:.3f}'
                ])
            table.add_row([
                'MEAN',
                sum(cls_numbers),
                f'{precision.mean().item():.3f}',
                f'{recall.mean().item():.3f}',
                f'{f1score.mean().item():.3f}',
                '-'
            ])
            logger.log('\n' + str(table))
        else:
            table = PrettyTable(['Class', 'Samples', 'Precision', 'Recall', 'F1-Score', 'Threshold'])
            
            for i, c in dataloader.dataset.id2label.items():
                table.add_row([
                    c, 
                    cls_numbers[i], 
                    f'{precision[i].item():.3f}',
                    f'{recall[i].item():.3f}',
                    f'{f1score[i].item():.3f}',
                    f'{thresh[i].item():.3f}' if isinstance(thresh, torch.Tensor) else f'{thresh:.3f}'
                ])
            
            table.add_row([
                'MEAN',
                sum(cls_numbers),
                f'{precision.mean().item():.3f}',
                f'{recall.mean().item():.3f}',
                f'{f1score.mean().item():.3f}',
                '-'
            ])
            
            # 显示表格
            logger.console('\n' + str(table))

    if pbar:
        if is_single_label:
            pbar.desc = f'{pbar.desc[:-36]}{loss:>12.3g}{top1:>12.3g}{top5:>12.3g}'
        else:
            pbar.desc = f'{pbar.desc[:-36]}{loss:>12.3g}{precision.mean().item():>12.3g}{recall.mean().item():>12.3g}{f1score.mean().item():>12.3g}'

    # filename = 'train_results.txt' if is_training else 'val_results.txt'
    # with open(filename, 'w') as f:
    #     if is_single_label:
    #         # Save each sample's prediction and true label
    #         for i in range(len(targets)):
    #             f.write(f'Sample {i}: GT={targets[i].item()}, Pred={pred[i, 0].item()}, Top{top_k}={[pred[i, j].item() for j in range(top_k)]}\n')
    #     else:
    #         # Save each sample's multi-label prediction and true label
    #         for i in range(len(targets)):
    #             gt = targets[i].cpu().numpy()
    #             pd = pred[i].cpu().numpy()
    #             gt_indices = np.where(gt == 1)[0]
    #             pred_indices = np.where(pd == 1)[0]
    #             f.write(f'Sample {i}:\n')
    #             f.write(f'  GT classes: {gt_indices.tolist()}\n')
    #             f.write(f'  Pred classes: {pred_indices.tolist()}\n')

    if lossfn:
        if is_single_label: 
            return top1, top5, loss
        else: 
            return precision.mean().item(), recall.mean().item(), f1score.mean().item(), loss
    else:
        if is_single_label: 
            return top1, top5
        else: 
            return precision.mean().item(), recall.mean().item(), f1score.mean().item()



import copy
from PIL import Image, ImageOps
from typing import Optional, List, Tuple, Callable, Union, Dict, Sequence
import numpy as np
import random
import torchvision.transforms as T
from functools import wraps
import torch.nn as nn
import glob
import os
from torchvision.transforms.functional import InterpolationMode
from abc import ABCMeta, abstractmethod
import cv2

# all methods based on PIL
__all__ = ['color_jitter', # 颜色抖动
           'random_color_jitter',# [随机]颜色抖动
           'random_horizonflip', # [随机]水平翻转
           'random_verticalflip', # [随机]上下翻转
           'random_doubleflip', # [随机]上下或左右翻转
           'random_crop', # [随机]抠图
           'random_augment', # RandAug
           'center_crop', # 中心抠图
           'resize', # 按最小边缩放
           'resize_and_padding', # 按最大边缩放+填充
           'centercrop_resize', # 中心抠图+缩放
           'random_cutout', # 随机CutOut
           'random_localgaussian', # 随机局部高斯
           'random_cutaddnoise', # 随机CutOut+增加噪音
           'random_affine', # 随机仿射变换
           'to_tensor', # 转Tensor
           'to_tensor_without_div', # 转Tensor不除255
           'normalize', # Normalize
           'random_gaussianblur', # 随机高斯模糊
           'random_autocontrast', # 随机对比度增强
           'random_adjustsharpness', # 随机锐化
           'random_rotate', # 随机(角度)旋转
           'random_invert', # 随机翻转 黑变白 白变黑 这种翻转
           'random_equalize',
           'random_augmix', # 随机样本自混合
           'random_grayscale', # 随机灰度 input几通道 forward也是几通道
           'random_crop_and_resize', # 随机crop再resize
           'pad2square', # 按最大边填充正方形
           'create_AugTransforms',
           'list_augments']

"""
References: https://pytorch.org/vision/stable/auto_examples/transforms/plot_transforms_illustrations.html#sphx-glr-auto-examples-transforms-plot-transforms-illustrations-py
"""

AUG_METHODS = {}
def register_method(fn: Callable):
    key = fn.__name__
    if key in AUG_METHODS:
        raise ValueError(f"An entry is already registered under the name '{key}'.")
    AUG_METHODS[key] = fn
    @wraps(fn)
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs)

    return wrapper

class Cutout:
    """Randomly mask out one or more patches from an image.
    Args:
        n_holes (int): Number of patches to cut out of each image.
        length (int): The length (in pixels) of each square patch.
    """
    def __init__(self, n_holes: int, length: int, ratio: float,
                 h_range: Optional[List[int]] = None, w_range: Optional[List[int]] = None,
                 prob: float = 0.5, color: tuple[int, int] = (0, 0)):
        self.n_holes = n_holes
        self.length = length
        self.ratio = ratio
        self.h_range = h_range
        self.w_range = w_range
        self.prob = prob
        self.color = color

    def __call__(self, image):
        """
        Args:
            img (Tensor): Tensor image of size (C, H, W) from PIL
        Returns:
            PIL: Image with n_holes of dimension length x length cut out of it.
        """
        if random.random() > self.prob:
            return image
        img = copy.deepcopy(image) # protect source image

        h = self.h_range if self.h_range is not None else [0, img.height] # PIL Image size->(w,h)
        w = self.w_range if self.w_range is not None else [0, img.width]

        mask_w = int(random.uniform(1-self.ratio, 1+self.ratio) * self.length)
        mask_h = self.length

        for n in range(self.n_holes):
            mask = Image.new('RGB', size=(mask_w, mask_h), color=(random.randint(*self.color), random.randint(*self.color), random.randint(*self.color)))
            # center
            y = np.random.randint(*h)
            x = np.random.randint(*w)

            # left-up
            x1 = max(0, x - self.length // 2)
            y1 = max(0, y - self.length // 2)

            img.paste(mask, (x1, y1))

        return  img

class CutAddNoise:
    """Randomly mask out one or more patches from an image.
    Args:
        n_holes (int): Number of patches to cut out of each image.
        length (int): The length (in pixels) of each square patch.
    """

    def __init__(self, n_holes: int, length: int, noisy_src: str,
                 h_range: Optional[List[int]] = None, w_range: Optional[List[int]] = None,
                 prob: float = 0.5, ):
        self.n_holes = n_holes
        self.length = length
        self.h_range = h_range
        self.w_range = w_range
        self.prob = prob
        self.noisy = glob.glob(f'{noisy_src}/*.jpg')
        assert os.path.splitext(self.noisy[0])[-1] == '.jpg', 'only support .jpg'

    def __call__(self, image):
        """
        Args:
            img (Tensor): Tensor image of size (C, H, W) from PIL
        Returns:
            PIL: Image with n_holes of dimension length x length cut out of it.
        """
        if random.random() > self.prob:
            return image
        img = copy.deepcopy(image)  # protect source image

        h = self.h_range if self.h_range is not None else [0, img.height]  # PIL Image size->(w,h)
        w = self.w_range if self.w_range is not None else [0, img.width]

        noisy_image = Image.open(random.choice(self.noisy)).convert('RGB')
        noisy_image = noisy_image.resize(size=(image.width, image.height))

        for n in range(self.n_holes):
            # center
            y = np.random.randint(*h)
            x = np.random.randint(*w)

            # left-up
            x1 = max(0, x - self.length // 2)
            y1 = max(0, y - self.length // 2)

            # right-bottom
            x2 = min(noisy_image.width, x + self.length // 2)
            y2 = min(noisy_image.height, y + self.length // 2)

            noisy_box = noisy_image.crop((x1, y1, x2, y2))
            img.paste(noisy_box, (x1, y1))

        return img

class CenterCropAndResize(nn.Sequential):
    def __init__(self, center_size, re_size):
        super().__init__(T.CenterCrop(center_size),
                         T.Resize(re_size, interpolation=InterpolationMode.BILINEAR))
        self.size = re_size

class RandomColorJitter(T.ColorJitter):
    def __init__(self, prob: float = 0.5, *args, **kargs):
        super().__init__(*args, **kargs)
        self.prob = prob

    def forward(self, img):
        r = random.random()
        if r < self.prob:
            return super().forward(img)
        else: return img

class PILToTensorNoDiv:
    def __init__(self):
        self.pil2tensor = T.PILToTensor()

    def __call__(self, pic):
        return self.pil2tensor(pic).float()

class BaseClassWiseAugmenter(metaclass=ABCMeta):
    def __init__(self, base_transforms: Dict, class_transforms_mapping: Optional[Dict[str, List[int]]]):
        self.base_transforms = create_AugTransforms(base_transforms)
        if class_transforms_mapping is not None:
            class_transforms = dict()
            for c, t in class_transforms_mapping.items():
                if isinstance(t, str): t = t.split()
                transform = []
                for i in t:
                    transform.append(self.base_transforms.transforms[int(i)])
                class_transforms[c] = T.Compose(transform)
            self.class_transforms = class_transforms
        else:
            self.class_transforms = None

    @abstractmethod
    def __call__(self, image, label: Union[List, int], class_indices: List[int]):
        return self.base_transforms(img=image)

class PadIfNeed:
    def __init__(self, pad_value: Union[int, Sequence], mode: str):
        if isinstance(pad_value, int):
            pad_value = (pad_value, pad_value, pad_value)
        else:
            assert len(pad_value) == 3, 'pad_value can only be a three-dimensional vector or an int'

        assert mode in ('edge', 'average'), "mode can only be 'edge' [fill one side] and 'average' [fill both sides]"

        self.pad_value = pad_value
        self.mode = mode

    def __call__(self, image):
        w, h = image.size
        max_size = max(w, h)
        new_im = Image.new('RGB', (max_size, max_size), self.pad_value)
        if self.mode == 'average':
            new_im.paste(image, ((max_size - w) // 2, (max_size - h) // 2))
        else:
            new_im.paste(image, (max_size-w, max_size-h))
        return new_im

class Reverse_PadIfNeed:
    def __init__(self, mode: str):
        self.mode = mode

    # image: square
    def __call__(self, image: np.ndarray, dsize: Tuple[int, int]):
        if dsize[0] == dsize[1]: return image

        dst_h, dst_w = dsize
        height, width = image.shape[:2]
        if self.mode == 'average':
            x1, y1 = (width-dst_w) // 2, (height-dst_h) // 2
            x2, y2 = x1 + dst_w, y1 + dst_h
        else:
            x1, y1 = width-dst_w, height-dst_h
            x2, y2 = x1 + dst_w, y1 + dst_h

        return image[x1: x2, y1: y2]

class LocalGaussian:
    def __init__(self, prob: float, ksize: Tuple[int, int], h_range: Optional[Tuple[int, int]] = None,
                 w_range: Optional[Tuple[int, int]] = None):
        """
        Argument:
            ksize: gaussian2D kernel size
            h_range: range of valid point
            w_range: range of valid point
        """
        self.prob = prob
        self.kszie = ksize
        self.valid_h_range = h_range
        self.valid_w_range = w_range

    def generate_seamless_mask(self, flaw: np.ndarray):
        mask = np.zeros(flaw.shape[:-1], dtype=np.uint8)
        if random.random() < 0.5:  # square
            mask.fill(255)
        else:  # ellipse
            # start_angle = random.randint(0, 180)
            # end_angle = max(start_angle + random.randint(90, 180), 360)
            h, w = mask.shape
            mask = cv2.ellipse(mask, (w // 2, h // 2), (w // 2, h // 2), 0, 0, 360, 255, -1)
        return mask

    def random_binary_mask_location(self, image: np.ndarray, local_height_range: Optional[Tuple[int, int]] = None,
                                    local_width_range: Optional[Tuple[int, int]] = None) -> Tuple[int, int, int, int]:

        assert type(image) is np.ndarray, 'image 必须是numpy.ndarray'
        height, width = image.shape if image.ndim == 2 else image.shape[:-1]
        if self.valid_w_range is None: self.valid_w_range = (0, width)
        if self.valid_h_range is None: self.valid_h_range = (0, height)

        for _ in range(10000):
            x = random.randint(*self.valid_w_range)
            y = random.randint(*self.valid_h_range)
            h = random.randint(*local_height_range if local_height_range is not None else (height // 5, height // 10 * 3))
            w = random.randint(*local_width_range if local_width_range is not None else (width // 5, width // 10 * 3))
            if x + w > self.valid_w_range[1] or y + h > self.valid_h_range[1]:
                continue
            else:
                break

        return x, y, w, h

    def __call__(self, image):
        if random.random() < self.prob:
            array_image = cv2.cvtColor(np.asarray(image), cv2.COLOR_RGB2BGR)
            blur = cv2.GaussianBlur(array_image, self.kszie, sigmaX=0)
            x, y, w, h = self.random_binary_mask_location(blur)

            width, height = blur.shape[1], blur.shape[0]
            x = min(x, width - w)
            y = min(y, height - h)

            roi = blur[y:y + h, x:x + w]

            mask = self.generate_seamless_mask(roi)

            try:
                merge_image = cv2.seamlessClone(roi, array_image, mask, p=(x + w // 2, y + h // 2), flags=cv2.NORMAL_CLONE)
            except cv2.error:
                return image

            return Image.fromarray(cv2.cvtColor(merge_image, cv2.COLOR_BGR2RGB))
        else:
            return image

class RandomDoubleFlip:
    def __init__(self, prob: Union[Sequence, float] = 0.5):
        assert type(prob) in (Sequence, float), 'prob should Sequence or float'
        self.prob = prob if isinstance(prob, Sequence) else (prob, prob)
        self.choices = [random_verticalflip(0.5), random_horizonflip(0.5)]

    def __call__(self, image):
        return random.choices(self.choices, weights=self.prob, k=1)[0](image)

class ResizeAndPadding2Square:

    """
    size is an int, longger edge of the image will be matched to this number.
    This function is resize as the longger edge and padding 0, put the image as center.
    """

    def __init__(self, size: int = 224, training: bool = False) -> None:
        self.size = size
        self.training = training

    def __call__(self, image):
        if self.training:
            resample = Image.BILINEAR if random.random() < 0.5 else Image.NEAREST
        else: 
            resample = Image.BILINEAR
    
        # Get original image dimensions
        width, height = image.size
        max_side = max(width, height)

        # Scale image based on the larger side
        scale_factor = self.size / max_side
        new_width = int(width * scale_factor)
        new_height = int(height * scale_factor)

        # Resize the image while maintaining aspect ratio
        image = image.resize((new_width, new_height), resample)
        
        # Calculate padding to make the image square and centered
        pad_width = (self.size - new_width) // 2
        pad_height = (self.size - new_height) // 2
        
        # Apply padding to center the image
        padding = (pad_width, pad_height, self.size - new_width - pad_width, self.size - new_height - pad_height)
        padded_image = ImageOps.expand(image, padding, fill=(0, 0, 0))  # You can change fill color if needed

        return padded_image
    
    def __repr__(self) -> str:
        return f"{super().__repr__()}(size={self.size})"

class ReverseResizeAndPadding2Square:
    def __init__(self, size: int = 224) -> None:
        self.size = size

    def __call__(self, image: np.ndarray, dsize: Tuple[int, int]) -> np.ndarray:
        # Get original image dimensions
        width, height = dsize
        max_side = max(width, height)

        # Scale image based on the larger side
        scale_factor = self.size / max_side
        new_width = int(width * scale_factor)
        new_height = int(height * scale_factor)

        pad_width = (self.size - new_width) // 2
        pad_height = (self.size - new_height) // 2

        image = image[pad_height: pad_height + new_height, pad_width: pad_width + new_width]

        image = cv2.resize(image, dsize, cv2.INTER_LINEAR)

        return image

class RandomResizedCrop(T.RandomResizedCrop):
    def __init__(self, size, scale=(0.08, 1.0), ratio=(3./4., 4./3.), interpolation=InterpolationMode.BILINEAR, antialias: Optional[bool] = True):
        super().__init__(size, scale, ratio, interpolation, antialias)
        self.resize_and_padding = ResizeAndPadding2Square(size=size, training=True)
    
    def forward(self, img):
        w, h = img.size
        ratio = max(h/w, w/h)
        if ratio > 1.5:
            return self.resize_and_padding(img)
        else: return super().forward(img)
    
@register_method
def random_cutout(n_holes:int = 1, length: int = 200, ratio: float = 0.2,
                  h_range: Optional[List[int]] = None, w_range: Optional[List[int]] = None, prob: float = 0.5, color: tuple[int, int] = (0, 0)):
    return Cutout(n_holes, length, ratio, h_range, w_range, prob, color)

@register_method
def random_cutaddnoise(n_holes:int = 1, length: int = 200, noisy_src: str = None,
                  h_range: Optional[List[int]] = None, w_range: Optional[List[int]] = None, prob: float = 0.5):
    return CutAddNoise(n_holes, length, noisy_src, h_range, w_range, prob)

@register_method
def random_localgaussian(prob: float = 0.5, ksize: Tuple[int, int] = (7, 7), *args, **kwargs):
    return LocalGaussian(prob=prob, ksize=ksize, *args, **kwargs)

@register_method
def color_jitter(brightness: float = 0.1,
                 contrast: float = 0.1,
                 saturation: float = 0.1,
                 hue: float = 0.1):
    return T.ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue)

@register_method
def random_autocontrast(p: float=0.5):
    return T.RandomAutocontrast(p=p)

@register_method
def random_adjustsharpness(sharpness_factor: float=2, p=0.5):
    return T.RandomAdjustSharpness(sharpness_factor, p=p)

@register_method
def random_invert(p: float=0.5):
    return T.RandomInvert(p=p)

@register_method
def random_equalize(p: float=0.5):
    return T.RandomEqualize(p=p)

@register_method
def random_augmix(*args, **kwargs):
    return T.AugMix(*args, **kwargs)

@register_method
def random_crop(*args, **kwargs):
    return T.RandomCrop(*args, **kwargs)
@register_method
def random_color_jitter(prob: float = 0.5, *args, **kwargs):
    # brightness: float = 0.1, contrast: float = 0.1, saturation: float = 0.1, hue: float = 0.1
    return RandomColorJitter(prob = prob, *args, **kwargs)

@register_method
def random_horizonflip(p: float = 0.5):
    return T.RandomHorizontalFlip(p=p)

@register_method
def random_verticalflip(p: float = 0.5):
    return T.RandomVerticalFlip(p=p)

@register_method
def random_doubleflip(prob: Union[Sequence, float] = 0.5):
    return RandomDoubleFlip(prob = prob)

@register_method
def random_rotate(degrees: Union[Sequence, int]):
    return T.RandomRotation(degrees = degrees, interpolation=InterpolationMode.BILINEAR)
@register_method
def to_tensor():
    return T.ToTensor()

@register_method
def to_tensor_without_div():
    return PILToTensorNoDiv()

@register_method
def normalize(mean: Sequence = (0.485, 0.456, 0.406), std: Sequence = (0.229, 0.224, 0.225)):
    return T.Normalize(mean=mean if isinstance(mean, Sequence) else eval(mean),
                       std=std if isinstance(std, Sequence) else eval(std))

@register_method
def random_augment(num_ops: int = 2, magnitude: int = 9, num_magnitude_bins: int = 31,):
    return T.RandAugment(num_ops=num_ops, magnitude=magnitude, num_magnitude_bins=num_magnitude_bins)

@register_method
def center_crop(size):
    # size (sequence or int): Desired output size of the crop. If size is an
    # int instead of sequence like (h, w), a square crop (size, size) is
    # made. If provided a sequence of length 1, it will be interpreted as (size[0], size[0]).
    return T.CenterCrop(size=size)

@register_method
def resize(size = 224):
    # size (sequence or int) -> square or rectangle: Desired output size. If size is a sequence like
    # (h, w), output size will be matched to this. If size is an int,smaller
    # edge of the image will be matched to this number. i.e,
    # if height > width, then image will be rescaled to (size * height / width, size).
    return T.Resize(size = size, interpolation=InterpolationMode.BILINEAR)

@register_method
def resize_and_padding(size: int = 224, training: bool = False):
    return ResizeAndPadding2Square(size=size, training = training)

@register_method
def centercrop_resize(center_size: tuple, re_size: tuple):
    return CenterCropAndResize(center_size, re_size)

@register_method
def random_affine(degrees = 0., translate = 0., scale = 0., shear = 0., fill=0, center=None):
    return T.RandomAffine(degrees=degrees, translate=translate, scale=scale, shear=shear, fill=fill, center=center)

@register_method
def random_gaussianblur(prob: float = 0.5, kernel_size=3, sigma=(0.1, 2.0)): # 每次transform sigma会均匀采样一次 除非传sigma是固定值
    return T.RandomApply([T.GaussianBlur(kernel_size=kernel_size, sigma=sigma)], p = prob)

@register_method
def random_grayscale(p: float = 0.5): # 图是几通道 灰度输出也是几通道
    return T.RandomGrayscale(p=p)

@register_method
def random_crop_and_resize(size, *args, **kwargs):
    return RandomResizedCrop(size = size, *args, **kwargs)

@register_method
def pad2square(pad_value: Union[int, Sequence] = 0, mode: str = 'average'):
    return PadIfNeed(pad_value, mode)

@register_method
def random_choice(transforms: list, p: Optional[Union[Sequence, float]] = None):
    return T.RandomChoice(transforms=transforms, p = p)

def create_AugTransforms(augments: list):

    def addAugToSequence(aug_name: str, params: Union[dict, str], aug_list: list) -> None:
        if params == 'no_params':
            aug_list.append(AUG_METHODS[aug_name]())
        else:
            assert isinstance(params, dict), 'params must be passed as key-value pairs, such as dict'
            aug_list.append(AUG_METHODS[aug_name](**params))

    augs = []
    for aug in augments:
        for key, params in aug.items():
            if key == 'random_choice':
                assert "transforms" in params and isinstance(params["transforms"], list), 'random_choice should have "transforms" keys'
                choice_aug_list = []
                for choice in params["transforms"]:
                    assert isinstance(choice, dict) and len(choice)==1, f'every augment methord must be dict in random_choice, {len(params)}augments need to be {len(params)} dicts'
                    choice_key, choice_param = tuple(*choice.items())
                    addAugToSequence(choice_key, choice_param, choice_aug_list)
                # 把random_choice作为单独的aug加进去
                kwargs = {"transforms": choice_aug_list, "p": params.get("p", None)}
                augs.append(AUG_METHODS[key](**kwargs))
            else:
                addAugToSequence(key, params, augs)

    return T.Compose(augs)
    # augments = augments.strip().split()
    # return T.Compose(tuple(map(lambda x: AUG_METHODS[x](**kwargs) if x not in _imgsz_related_methods else AUG_METHODS[x](imgsz, **kwargs), augments)))

def list_augments():
    augments = [k for k, v in AUG_METHODS.items()]
    return sorted(augments)

# transforms about image size
SPATIAL_TRANSFORMS = set([T.CenterCrop, T.Resize, CenterCropAndResize, T.RandomCrop, T.RandomResizedCrop, PadIfNeed, ResizeAndPadding2Square])


from .models import *
from .engine import *
from .utils import *
from .dataset import *
from .distills import *
from .engine.procedure import valuate as valuate_classifier
from .engine.representation import valuate_cbir, valuate_face



from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup(
    name='doraemon-torch', 
    version='0.0.09a',
    author='duke',
    author_email='dk812821001@163.com',
    description='Doraemon',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/wuji3/Doraemon',
    packages=find_packages(include=['doraemon', 'doraemon.*']),
    python_requires='>=3.10',
    install_requires=[
        'numpy>=1.24.3',
        'torch>=2.5.1',
        'torchvision>=0.20.0',
        'torchaudio>=2.5.1',
        'transformers>=4.48.3',
        'torchmetrics>=0.11.4',
    ],
    extras_require={
        'deploy': [
            'Pillow>=9.4.0',
            'opencv-python>=4.7.0.72',
        ],
        'full': [
            'Pillow>=9.4.0',
            'opencv-python>=4.7.0.72',
            'tqdm>=4.66.4',
            'grad-cam>=1.4.8',
            'timm>=0.9.16',
            'tensorboard>=2.16.2',
            'prettytable>=3.10.0',
            'datasets>=2.20.0',
            'imagehash>=4.3.1',
            'faiss-cpu>=1.7.2',
        ],
    },
)


import torch
from torchvision.transforms import CenterCrop, Resize, Compose, RandomChoice
from ..dataset.transforms import ResizeAndPadding2Square, RandomResizedCrop
import torch.nn as nn
from torch.cuda.amp import GradScaler
from ..dataset.basedataset import ImageDatasets
from ..dataset.transforms import CenterCropAndResize, SPATIAL_TRANSFORMS, create_AugTransforms
from torch.utils.data import DistributedSampler
from ..utils.logger import SmartLogger
from .optimizer import create_Optimizer
from .scheduler import create_Scheduler
from ..models.losses.loss import create_Lossfn
from .procedure.train import Trainer
from functools import partial
from pathlib import Path
from torch.nn.parallel import DistributedDataParallel as DDP
from ..utils.plots import colorstr
from ..structure.sampler import OHEMImageSampler
from ..built.layer_optimizer import SeperateLayerParams
import time
import os
import datetime
import yaml
from copy import deepcopy
from ..models.ema import ModelEMA
from ..models import get_model
from ..dataset.dataprocessor import SmartDataProcessor
from ..utils.average_meter import AverageMeter
import torch.distributed as dist


__all__ = ['yaml_load', 'CenterProcessor','increment_path']


def yaml_load(file='data.yaml'):
    # Single-line safe yaml loading
    with open(file, errors='ignore') as f:
        return yaml.safe_load(f)


def increment_path(path, exist_ok=False, sep='', mkdir=False):
    # Increment file or directory path, i.e. runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc.
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        path, suffix = (path.with_suffix(''), path.suffix) if path.is_file() else (path, '')

        # Method 1
        for n in range(2, 9999):
            p = f'{path}{sep}{n}{suffix}'  # increment path
            if not os.path.exists(p):  #
                break
        path = Path(p)

    if mkdir:
        path.mkdir(parents=True, exist_ok=True)  # make directory

    return path

def get_imgsz(augment: dict):
    augments = create_AugTransforms(augment)
    for a in augments.transforms[::-1]:
        if type(a) in SPATIAL_TRANSFORMS and hasattr(a, 'size'):
            if type(a.size) is int: return (a.size, a.size)
            elif type(a.size) in [tuple, list]: return tuple(a.size)
            else: raise ValueError('size be int, tuple or list')

class CenterProcessor:
    def __init__(self, cfgs: dict, rank: int, project: str = None, train: bool = True, opt = None):
        log_filename = Path(project) / "log{}.log".format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S")) if project is not None and train else None
        self.project = project
        if rank in {-1, 0} and train:
            project.mkdir(parents=True, exist_ok=True)

        self.cfgs = cfgs
        self.model_cfg = cfgs['model']
        self.data_cfg = cfgs['data']
        self.hyp_cfg = cfgs['hyp']
        self.opt = opt
        self.imgsz = (cfgs['model']['image_size'], )

        # task
        self.task = self.model_cfg['task'] 

        # rank
        self.rank: int = rank
        # device
        if rank != -1:
            device = torch.device('cuda', rank)
        else:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.device: torch.device = device

        # logger
        self.logger = SmartLogger(filename=log_filename, level=1) if rank in {-1,0} else None
        if self.logger is not None and rank in {-1, 0} and train:
            self.logger.console(cfgs) # output configs

        # model processor
        self.model_processor = get_model(self.model_cfg, self.logger, rank)
        self.model_processor.model.to(device)
        # data processor
        self.data_processor = SmartDataProcessor(self.data_cfg, rank=rank, project=project, training = train)
        if self.task == 'classification' and train:
            self.data_processor.val_dataset = self.data_processor.create_dataset('val', training = train)

        # loss
        loss_choice: str = 'ce' if self.hyp_cfg['loss']['ce'] else 'bce'
        self.loss_choice = loss_choice
        if self.task == 'classification':
            if train:
                self.lossfn = create_Lossfn(loss_choice)() \
                    if loss_choice == 'bce' \
                    else create_Lossfn(loss_choice)(label_smooth = self.hyp_cfg['label_smooth'])
            self.thresh = self.hyp_cfg['loss']['bce'][1] if loss_choice == 'bce' else 0

            # add label_transforms
            if loss_choice == 'bce' and train:
                self.data_processor.train_dataset.label_transforms = \
                    partial(ImageDatasets.set_label_transforms,
                            num_classes = self.model_cfg['num_classes'],
                            label_smooth = self.hyp_cfg['label_smooth'])
                self.data_processor.val_dataset.label_transforms = \
                    partial(ImageDatasets.set_label_transforms,
                            num_classes=self.model_cfg['num_classes'],
                            label_smooth=0)
                # bce not support self.sampler
                self.sampler = None
            # ohem
            elif self.hyp_cfg['strategy']['ohem'][0]: self.sampler = OHEMImageSampler(*self.hyp_cfg['strategy']['ohem'][1:])
            else: self.sampler = None

        else: self.lossfn = create_Lossfn(loss_choice)(label_smooth = self.hyp_cfg['label_smooth'])

        if train and self.task == 'classification':
            # mixup sampler
            mixup_ratio, mixup_duration = self.hyp_cfg['strategy']['mixup']
            self.mixup_duration = mixup_duration
            self.mixup_ratio = mixup_ratio

            # progressive learning
            self.prog_learn = self.hyp_cfg['strategy']['prog_learn']

            if self.prog_learn: 
                warmup_epochs = self.hyp_cfg['warm_ep']
                remaining_epochs = self.hyp_cfg['epochs'] - warmup_epochs
                stage1_epochs = remaining_epochs // 4
                stage2_epochs = remaining_epochs // 4
                resize_chnodes = [
                    warmup_epochs, 
                    warmup_epochs + stage1_epochs,  
                    warmup_epochs + stage1_epochs + stage2_epochs
                ]
                self.resize_chnodes = resize_chnodes

                min_imgsz = min(self.imgsz)
                self.imgsz_milestone = torch.linspace(int(min_imgsz * 0.5), int(min_imgsz), 3, dtype=torch.int32).tolist()

            # focalloss hard
            if loss_choice == 'bce' and self.hyp_cfg['strategy']['focal'][0]:
                self.focal = create_Lossfn('focal')(gamma=self.hyp_cfg['strategy']['focal'][2], alpha= self.hyp_cfg['strategy']['focal'][1])
            else:
                self.focal = None

        # ema
        if train: self.ema = ModelEMA(self.model_processor.model) if rank in {-1, 0} else None

        self.loss_meter = AverageMeter()

    def set_optimizer_momentum(self, momentum) -> None:
        for g in self.optimizer.param_groups:
            g['momentum'] = momentum

    def auto_mixup(self, mixup: float, epoch:int, milestone: list) -> float:
        if mixup == 0 or epoch < milestone[0] or self.dist_sampler['beta'] is None : return 0
        else:
            mix_prob = self.dist_sampler['uniform'].sample()
            lam = self.dist_sampler['beta'].sample().to(self.device) if mix_prob < mixup else 0

            return lam

    def auto_prog(self, epoch: int):
        def create_AugSequence(train_augs : list, size):
            sequence = []
            for i, m in enumerate(train_augs):
                if isinstance(m, RandomChoice):
                    m.transforms = create_AugSequence(m.transforms, size)
                    sequence.append(m)
                elif isinstance(m, ResizeAndPadding2Square):
                    m.size = size
                    sequence.append(m)
                elif isinstance(m, CenterCrop):
                    if i + 1 < len(train_augs) and (not isinstance(train_augs[i + 1], Resize) or not isinstance(train_augs[i + 1], RandomResizedCrop)):
                        sequence.extend([m, Resize(size)])
                    else:
                        sequence.append(m)
                elif isinstance(m, Resize):
                    sequence.append(Resize(size))
                elif isinstance(m, CenterCropAndResize):
                    m[-1] = Resize(size)
                    sequence.append(m)
                elif isinstance(m, RandomResizedCrop):
                    m.size = (size, size)
                    m.resize_and_padding.size = size
                    sequence.append(m)
                else:
                    sequence.append(m)

            return sequence

        chnodes = self.resize_chnodes

        if epoch == chnodes[0]: size = self.imgsz_milestone[0]
        elif epoch == chnodes[1]: size = self.imgsz_milestone[1]
        elif epoch == chnodes[2]: size = self.imgsz_milestone[2]
        else: return

        if hasattr(self.data_processor.train_dataset.transforms, 'base_transforms'):
            transforms = self.data_processor.train_dataset.transforms.base_transforms.transforms
            self.data_processor.train_dataset.transforms.base_transforms = Compose(create_AugSequence(transforms, size))
        if hasattr(self.data_processor.train_dataset.transforms, 'class_transforms') and self.data_processor.train_dataset.transforms.class_transforms is not None:
            for c, transforms in self.data_processor.train_dataset.transforms.class_transforms.items():
                self.data_processor.train_dataset.transforms.class_transforms[c] = Compose(create_AugSequence(transforms.transforms, size))

    def set_sync_bn(self):
        self.model_processor.model = nn.SyncBatchNorm.convert_sync_batchnorm(module=self.model_processor.model)

    def run_classifier(self, resume = None): # train+val per epoch
        last, best = self.project / 'last.pt', self.project / 'best.pt'

        model, data_processor, scaler, device, epochs, logger, mixup_duration, rank, warm_ep, aug_epoch, focal, sampler, thresh = \
            self.model_processor.model, self.data_processor, \
            GradScaler(enabled = (self.device != torch.device('cpu'))), self.device, self.hyp_cfg['epochs'], \
            self.logger, self.mixup_duration, self.rank, self.hyp_cfg['warm_ep'], \
            self.data_cfg['train']['aug_epoch'], self.focal, self.sampler, self.thresh

        # data
        train_dataset, val_dataset = data_processor.train_dataset, data_processor.val_dataset
        data_sampler = None if self.rank == -1 else DistributedSampler(dataset=train_dataset)
        train_dataloader = data_processor.set_dataloader(dataset=train_dataset,
                                                         bs=self.data_cfg['train']['bs'],
                                                         nw=self.data_cfg['nw'],
                                                         pin_memory=True,
                                                         sampler=data_sampler,
                                                         shuffle=data_sampler is None,
                                                         collate_fn=train_dataset.collate_fn,
                                                         drop_last = True,
                                                         prefetch_factor=2,
                                                         persistent_workers=True)

        if self.rank in {-1, 0}:
            val_dataloader = data_processor.set_dataloader(dataset=val_dataset,
                                                           bs=self.data_cfg['val']['bs'],
                                                           nw=self.data_cfg['nw'],
                                                           pin_memory=False,
                                                           shuffle=False,
                                                           collate_fn=val_dataset.collate_fn)
        else:
            val_dataloader = None

        # tell data distribution
        if rank in (-1, 0):
            ImageDatasets.tell_data_distribution({"train": train_dataset, "val": val_dataset}, logger, self.model_cfg['num_classes'])

        # optimizer
        params = SeperateLayerParams(model)
        optimizer = create_Optimizer(optimizer=self.hyp_cfg['optimizer'][0],
                                     lr=self.hyp_cfg['lr0'],
                                     weight_decay=self.hyp_cfg['weight_decay'],
                                     momentum=self.hyp_cfg['warmup_momentum'],
                                     params=params.create_ParamSequence(layer_wise=self.hyp_cfg['optimizer'][1],
                                                                        lr=self.hyp_cfg['lr0']))
        self.optimizer = optimizer
        # scheduler
        scheduler = create_Scheduler(scheduler=self.hyp_cfg['scheduler'],
                                     optimizer=optimizer,
                                     warm_ep=self.hyp_cfg['warm_ep'],
                                     epochs=self.hyp_cfg['epochs'],
                                     lr0=self.hyp_cfg['lr0'],
                                     lrf_ratio=self.hyp_cfg['lrf_ratio'])

        best_fitness = 0.
        start_epoch = 0

        # resume
        if resume is not None:
            ckp = torch.load(resume, map_location=device)
            start_epoch = ckp['epoch'] + 1
            best_fitness = ckp['best_fitness']
            if self.rank in {-1, 0}:
                self.ema.ema.load_state_dict(ckp['ema'].float().state_dict())
                self.ema.updates = ckp['updates']
                self.cfgs = ckp['config']
            model.load_state_dict(ckp['model'])
            optimizer.load_state_dict(ckp['optimizer'])
            scheduler.load_state_dict(ckp['scheduler'])
            if device != torch.device('cpu'):
                scaler.load_state_dict(ckp['scaler'])

            if rank in (-1, 0): 
                logger.both(f'resume: {resume}')

        load_from = self.model_cfg.get('load_from', None)
        if load_from is not None:
            state_dict = torch.load(load_from, weights_only=False)
            if 'ema' in state_dict: 
                state_dict = state_dict['ema'].state_dict()
            else: 
                state_dict = state_dict['model_state_dict'].state_dict()
            missing_keys, unexpected_keys = model.load_state_dict(state_dict=state_dict, strict=False)
            if rank in (-1, 0): 
                logger.both(f'load_from: {load_from}')
                logger.both(f"Missing keys: {missing_keys}")
                logger.both(f"Unexpected keys: {unexpected_keys}")

        if rank != -1:
            model = DDP(model, device_ids=[self.rank])

        if self.rank in {-1, 0}:

            if thresh == 0:
                print(f"{'Epoch':>10}{'GPU_mem':>10}{'train_loss':>12}{f'val_loss':>12}{'top1_acc':>12}{'top5_acc':>12}")
            else:
                print(f"{'Epoch':>10}{'GPU_mem':>10}{'train_loss':>12}{f'val_loss':>12}{'precision':>12}{'recall':>12}{'f1score':>12}")
            time.sleep(0.2)

        # total epochs
        total_epoch = epochs

        # trainer
        trainer = Trainer(model=model, 
                          train_dataloader=train_dataloader, 
                          val_dataloader=val_dataloader, 
                          optimizer=optimizer,
                          scaler=scaler, 
                          device=device, 
                          epochs=total_epoch, 
                          logger=logger, 
                          rank=rank, 
                          scheduler=scheduler, 
                          ema=self.ema, 
                          sampler=sampler, 
                          thresh=thresh,
                          teacher=self.teacher if hasattr(self, 'teacher') else None, 
                          cfgs=self.cfgs)

        t0 = time.time()
        for epoch in range(start_epoch, total_epoch):
            # warmup set augment as val
            if epoch == 0:
                self.data_processor.set_augment('train', transforms=None)
                trainer.mixup_sampler = None

            # change optimizer momentum from warm_moment0.8 -> momentum0.937
            if epoch == warm_ep:
                self.set_optimizer_momentum(self.hyp_cfg['momentum'])
                self.data_processor.set_augment('train', transforms=create_AugTransforms(self.data_cfg['train']['augment']))
                if self.mixup_ratio == 0 or self.mixup_duration == 0:
                    trainer.mixup_sampler = None
                else:
                    trainer.mixup_sampler = torch.distributions.beta.Beta(self.mixup_ratio, self.mixup_ratio)
                    if rank in {-1, 0}:
                        logger.both(f"Mixup start up")
            
            if (self.mixup_ratio != 0 and self.mixup_duration != 0) and epoch == warm_ep + mixup_duration:
                trainer.mixup_sampler = None
                if rank in {-1, 0}:
                    logger.both(f"Mixup end")

            # change lossfn bce -> focal
            if epoch == warm_ep and self.focal is not None:
                self.lossfn = self.focal
            
            # weaken data augment at milestone
            self.data_processor.auto_aug_weaken(int(epoch), milestone=aug_epoch)

            # progressive learning: effect on imagesz & mixup
            if self.prog_learn:
                self.auto_prog(epoch=epoch)

            # train for one epoch
            fitness = trainer.train_one_epoch(epoch, self.lossfn)

            if rank in {-1, 0}:
                # Best fitness
                if fitness > best_fitness:
                    best_fitness = fitness

                # Save model
                final_epoch: bool = epoch + 1 == total_epoch
                ckpt = {
                    'epoch': epoch,
                    'best_fitness': best_fitness,
                    'model': model.state_dict() if rank == -1 else model.module.state_dict(),  # deepcopy(de_parallel(model)).half(),
                    'ema': deepcopy(self.ema.ema),
                    'updates': self.ema.updates,
                    'optimizer': optimizer.state_dict(),  # optimizer.state_dict(),
                    'scheduler': scheduler.state_dict(),
                    'config': self.cfgs,
                    'id2label': train_dataset.id2label,
                    'label2id': train_dataset.label2id
                }
                if device != torch.device('cpu'):
                    ckpt['scaler'] = scaler.state_dict()

                # Save last, best and delete
                torch.save(ckpt, last)
                if best_fitness == fitness:
                    torch.save(ckpt, best)
                del ckpt

                # Training complete
                if final_epoch:
                    dataset = data_processor.train_dataset
                    
                    # Determine data path based on dataset type
                    data_path = (os.path.join(data_processor.data_cfgs["root"], 'val') 
                               if dataset.is_local_dataset and not dataset.multi_label
                               else data_processor.data_cfgs["root"])
                    
                    # Build base predict command
                    predict_cmd = (f'python -m scripts.infer {best} '  # model-path argument
                                 f'--data {data_path} '
                                 f'--split val '
                                 f'--ema '
                                 )
                    
                    # Build validation command
                    validate_cmd = (f'python -m scripts.validate {best} --ema')
                    if self.hyp_cfg['loss']['ce']: 
                        validate_cmd += " --eval_topk 5"

                    logger.both(f'\nTraining complete ({(time.time() - t0) / 3600:.3f} hours)'
                              f"\nResults saved to {colorstr('bold', self.project)}"
                              f'\nPredict:         {predict_cmd}'
                              f'\n             └── Optional: --classes {colorstr("blue", "<YOUR_TARGET_CLASSES> nargs = + ")}'
                              f'\n             └── Optional: --infer-option default  # default: Infer + Visualize + CaseAnalysis, autolabel: Infer + Label'
                              f'\n             └── Optional: --sampling N             # Visualize N random samples'
                              f'\nValidate:        {validate_cmd}')
        if torch.distributed.is_initialized():
            dist.destroy_process_group()

    def run_embedding(self, resume = None):
        model, data_processor, scaler, device, epochs, logger, rank, warm_ep, aug_epoch, task = self.model_processor.model, self.data_processor, \
            GradScaler(enabled = (self.device != torch.device('cpu'))), self.device, self.hyp_cfg['epochs'], self.logger, self.rank, self.hyp_cfg['warm_ep'], \
            self.data_cfg['train']['aug_epoch'], self.model_cfg['task']

        # load for fine-tune
        load_from = self.model_cfg.get('load_from', None)
        if load_from is not None:
            state_dict = torch.load(load_from, weights_only=False)
            if 'ema' in state_dict: state_dict = state_dict['ema']
            else: 
                state_dict = state_dict['model_state_dict']
            missing_keys, unexpected_keys = model.trainingwrapper['backbone'].load_state_dict(state_dict=state_dict, strict=False)
            if rank in (-1, 0): 
                logger.both(f'load_from: {load_from}')
                logger.both(f"Missing keys: {missing_keys}")
                logger.both(f"Unexpected keys: {unexpected_keys}")

        # data
        train_dataset = data_processor.train_dataset
        data_sampler = None if self.rank == -1 else DistributedSampler(dataset=train_dataset)
        train_dataloader = data_processor.set_dataloader(dataset=train_dataset,
                                                         bs=self.data_cfg['train']['bs'],
                                                         nw=self.data_cfg['nw'],
                                                         pin_memory=True,
                                                         sampler=data_sampler,
                                                         shuffle=data_sampler is None,
                                                         collate_fn=train_dataset.collate_fn,
                                                         drop_last = True,
                                                         prefetch_factor=2,
                                                         persistent_workers=True)
        # tell data distribution
        if self.rank in (-1, 0):
            ImageDatasets.tell_data_distribution({"train": train_dataset}, logger, self.model_cfg['head'][next(iter(self.model_cfg['head'].keys()))]['num_class'])


        # optimizer
        params = SeperateLayerParams(model)
        optimizer = create_Optimizer(optimizer=self.hyp_cfg['optimizer'][0],
                                     lr=self.hyp_cfg['lr0'],
                                     weight_decay=self.hyp_cfg['weight_decay'],
                                     momentum=self.hyp_cfg['warmup_momentum'],
                                     params=params.create_ParamSequence(layer_wise=self.hyp_cfg['optimizer'][1],
                                                                        lr=self.hyp_cfg['lr0']))
        self.optimizer = optimizer
        # scheduler
        scheduler = create_Scheduler(scheduler=self.hyp_cfg['scheduler'],
                                     optimizer=optimizer,
                                     warm_ep=self.hyp_cfg['warm_ep'] * len(train_dataloader),
                                     epochs=self.hyp_cfg['epochs'] * len(train_dataloader),
                                     lr0=self.hyp_cfg['lr0'],
                                     lrf_ratio=self.hyp_cfg['lrf_ratio'])

        start_epoch = 0

        # resume
        if resume is not None:
            ckp = torch.load(resume, map_location=device)
            start_epoch = ckp['epoch'] + 1

            if self.rank in {-1, 0}:
                self.ema.ema.load_state_dict(ckp['ema'].float().state_dict())
                self.ema.updates = ckp['updates']
            model.load_state_dict(ckp['state_dict'])
            optimizer.load_state_dict(ckp['optimizer'])
            scheduler.load_state_dict(ckp['scheduler'])
            if device != torch.device('cpu'):
                scaler.load_state_dict(ckp['scaler'])

            if rank in (-1, 0): logger.both(f'resume: {resume}')

        if rank != -1:
            model = DDP(model, device_ids=[self.rank])

        if self.rank in {-1, 0}: time.sleep(0.2)

        # total epochs
        total_epoch = epochs

        # trainer
        trainer = Trainer(model=model, 
                          train_dataloader=train_dataloader, 
                          val_dataloader=None, 
                          optimizer=optimizer,
                          scaler=scaler, 
                          device=device, 
                          epochs=total_epoch, 
                          logger=logger, 
                          rank=rank, 
                          scheduler=scheduler, 
                          ema=self.ema, 
                          sampler=None, 
                          thresh=None,
                          teacher=self.teacher if hasattr(self, 'teacher') else None, 
                          task=task, 
                          print_freq=self.opt.print_freq, 
                          save_freq=self.opt.save_freq, 
                          cfgs=self.cfgs, 
                          out_dir=self.opt.save_dir)

        t0 = time.time()
        for epoch in range(start_epoch, total_epoch):
            # warmup set augment as val
            if epoch == 0:
                self.data_processor.set_augment('train', transforms=create_AugTransforms(self.data_cfg['val']['augment']))

            # change optimizer momentum from warm_moment0.8 -> momentum0.937
            if epoch == warm_ep:
                self.set_optimizer_momentum(self.hyp_cfg['momentum'])
                self.data_processor.set_augment('train', transforms=create_AugTransforms(self.data_cfg['train']['augment']))

            # weaken data augment at milestone
            self.data_processor.auto_aug_weaken(epoch, milestone=aug_epoch, sequence = create_AugTransforms(self.data_cfg['val']['augment']))

            # train for one epoch
            trainer.train_one_epoch_face(self.lossfn, epoch, self.loss_meter)

        if rank in (-1, 0):
            predict_cmd = (f'python -m scripts.infer {os.path.join(self.project, "Epoch_?.pt")} ')

            logger.both(f'\nTraining complete ({(time.time() - t0) / 3600:.3f} hours)'
                        f"\nResults saved to {colorstr('bold', self.project)}"
                        f'\nPredict:         {predict_cmd}'
                        f'\nValidate:        python -m scripts.validate '
                        f'{os.path.join(self.project, "Epoch_?.pt")} '
                        f'--ema')

        if torch.distributed.is_initialized():
            dist.destroy_process_group()
